@@ -16,8 +16,7 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-from .core import JaxTuple, lattice_join
-from .interpreters.partial_eval import Primitive
+from .core import JaxTuple, lattice_join, Primitive, AbstractTuple
 from .tree_util import register_pytree_node
 from .util import safe_map
 
@@ -53,6 +52,15 @@ jaxval_zeros_likers = {}
 jaxval_zeros_likers[JaxTuple] = zeros_like_impl_jaxtuple
 
 
+def zeros_like_aval(aval):
+  return aval_zeros_likers[type(aval)](aval)
+aval_zeros_likers = {}
+
+def zeros_like_abstract_tuple(tup):
+  return JaxTuple(map(zeros_like_aval, tup))
+aval_zeros_likers[AbstractTuple] = zeros_like_abstract_tuple
+
+
 def zeros_like_jaxval(val):
   return zeros_like_p.bind(val)
 
