@@ -31,6 +31,7 @@ def wraps(wrapped, fun, namestr="{fun}", docstr="{doc}", **kwargs):
     fun.__name__ = namestr.format(fun=get_name(wrapped))
     fun.__module__ = get_module(wrapped)
     fun.__doc__ = docstr.format(fun=get_name(wrapped), doc=get_doc(wrapped), **kwargs)
+    fun.__wrapped__ = wrapped
   finally:
     return fun
 
@@ -38,11 +39,17 @@ def get_name(fun): return getattr(fun, "__name__", "<unnamed function>")
 def get_module(fun): return getattr(fun, "__module__", "<unknown module>")
 def get_doc(fun): return getattr(fun, "__doc__", "")
 
+@transformation_with_aux
+def pytree_fun_to_jaxtupletree_fun(args_trees, *args):
+  py_args = map(build_tree, args_trees, args)
+  ans = yield py_args, {}
+  yield pytree_to_jaxtupletree(ans)
 
 @transformation_with_aux
-def pytree_fun_to_jaxtupletree_fun(in_trees, *args):
-  py_args = map(build_tree, in_trees, args)
-  ans = yield py_args
+def pytree_fun_to_jaxtupletree_fun2(kwargs_tree, args_trees, kwargs, *args):
+  py_args = map(build_tree, args_trees, args)
+  py_kwargs = build_tree(kwargs_tree, kwargs)
+  ans = yield py_args, py_kwargs
   yield pytree_to_jaxtupletree(ans)
 
 def apply_jaxtree_fun(fun, io_tree, *py_args):
@@ -59,14 +66,29 @@ pytree_to_jaxtupletree = partial(process_pytree, pack)
 
 
 @transformation_with_aux
-def pytree_fun_to_flatjaxtuple_fun(in_tree, *args):
-  py_args = tuple(tree_unflatten(in_tree, args))
-  ans = yield py_args
-  yield maybe_pytree_to_maybe_jaxtuple(ans)
+def pytree_fun_to_flatjaxtuple_fun(in_trees, *args):
+  py_args = map(tree_unflatten, in_trees, args)
+  ans = yield py_args, {}
+  yield pytree_to_flatjaxtuple(ans)
 
-def maybe_pytree_to_maybe_jaxtuple(maybe_tree):
-  flat_ans, out_tree = tree_flatten(maybe_tree)
+@transformation_with_aux
+def flatten_fun(in_tree, *args_flat):
+  py_args, py_kwargs = tree_unflatten(in_tree, args_flat)
+  ans = yield py_args, py_kwargs
+  yield pytree_to_flatjaxtuple(ans)
+
+def pytree_to_flatjaxtuple(pytree):
+  flat, out_tree = tree_flatten(pytree)
+  return pack(flat), out_tree
+
+
+@transformation_with_aux
+def flatten_fun_leafout(in_tree, *args_flat):
+  # like flatten_fun but doesn't pack output leaves
+  py_args, py_kwargs = tree_unflatten(in_tree, args_flat)
+  ans = yield py_args, py_kwargs
+  flat_ans, out_tree = tree_flatten(ans)
   if out_tree is leaf:
-    return maybe_tree, leaf
+    yield ans, out_tree
   else:
-    return pack(flat_ans), out_tree
+    yield pack(flat_ans), out_tree
