@@ -16,51 +16,78 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
+import itertools as it
+
 from . import partial_eval as pe
-from . import xla
-from . import pxla
 from .. import core as core
-from ..core import JaxTuple, Trace, Tracer, new_master, get_aval, pack, call_p, Primitive
-from ..ad_util import (add_jaxvals, add_jaxvals_p, zeros_like_jaxval,
+from ..core import JaxTuple, Trace, Tracer, new_master, get_aval, pack, call_p, Primitive, Literal
+from ..ad_util import (add_jaxvals, add_jaxvals_p, zeros_like_jaxval, zeros_like_aval,
                        zeros_like_p, zero, Zero)
+from ..abstract_arrays import raise_to_shaped
 from ..util import unzip2, unzip3, safe_map, safe_zip, partial
-from ..tree_util import process_pytree, build_tree, register_pytree_node, prune
+from ..tree_util import process_pytree, build_tree, register_pytree_node, tree_map
 from ..linear_util import thunk, staged, transformation, transformation_with_aux, wrap_init
 
 from six.moves import builtins, reduce
 
 zip = safe_zip
 map = safe_map
+def identity(x): return x
 
-def jvp(fun):
-  return jvpfun(jvp_subtrace(fun))
+
+def jvp(fun, has_aux=False, instantiate=True):
+  if not has_aux:
+    return jvpfun(jvp_subtrace(fun), instantiate)
+  else:
+    fun, aux = jvp_subtrace_aux(fun, instantiate)
+    return jvpfun(fun, instantiate), aux
 
 @transformation
-def jvpfun(primals, tangents):
+def jvpfun(instantiate, primals, tangents):
   with new_master(JVPTrace) as master:
-    out_primal, out_tangent = yield master, primals, tangents
+    out_primal, out_tangent = yield (master, primals, tangents), {}
     del master
-  out_tangent = instantiate_zeros(out_primal, out_tangent)
+  out_tangent = instantiate_zeros_at(instantiate, out_primal, out_tangent)
   yield (out_primal, out_tangent)
 
+
 @transformation
 def jvp_subtrace(master, primals, tangents):
   trace = JVPTrace(master, core.cur_sublevel())
   for x in list(primals) + list(tangents):
     if isinstance(x, Tracer):
       assert x.trace.level < trace.level
-  ans = yield map(partial(JVPTracer, trace), primals, tangents)
+  ans = yield map(partial(JVPTracer, trace), primals, tangents), {}
   out_tracer = trace.full_raise(ans)
   out_primal, out_tangent = out_tracer.primal, out_tracer.tangent
   yield (out_primal, out_tangent)
 
+@transformation_with_aux
+def jvp_subtrace_aux(instantiate, master, primals, tangents):
+  trace = JVPTrace(master, core.cur_sublevel())
+  for x in list(primals) + list(tangents):
+    if isinstance(x, Tracer):
+      assert x.trace.level < trace.level
+  ans, aux = yield map(partial(JVPTracer, trace), primals, tangents), {}
+  out_tracer, aux_tracer = map(trace.full_raise, (ans, aux))
+  out_primal, out_tangent = out_tracer.primal, out_tracer.tangent
+  aux = aux_tracer.primal  # ignore aux tangent
+  out_tangent = instantiate_zeros_at(instantiate, out_primal, out_tangent)
+  yield (out_primal, out_tangent), aux
+
+
 @transformation
 def pack_output(*args):
-  ans = yield args
+  ans = yield args, {}
   yield pack(ans)
 
-def linearize(traceable, *primals):
-  jvpfun = pack_output(jvp(traceable))
+def linearize(traceable, *primals, **kwargs):
+  has_aux = kwargs.pop('has_aux', False)
+  if not has_aux:
+    jvpfun = pack_output(jvp(traceable))
+  else:
+    jvpfun, aux = jvp(traceable, has_aux=True)
+    jvpfun = pack_output(jvpfun)
   tangent_avals = [get_aval(p).at_least_vspace() for p in primals]
   in_pvals = (pe.PartialVal((None, pack(primals))),
               pe.PartialVal((core.AbstractTuple(tangent_avals), core.unit)))
@@ -68,17 +95,27 @@ def linearize(traceable, *primals):
   pval_primal, pval_tangent = unpair_pval(out_pval)
   aval_primal, const_primal = pval_primal
   assert aval_primal is None
-  return const_primal, pval_tangent, jaxpr, consts
+  if not has_aux:
+    return const_primal, pval_tangent, jaxpr, consts
+  else:
+    return const_primal, pval_tangent, jaxpr, consts, aux()
 
-def vjp(traceable, primals):
-  out_primal, pval, jaxpr, consts = linearize(traceable, *primals)
+def vjp(traceable, primals, has_aux=False):
+  if not has_aux:
+    out_primal, pval, jaxpr, consts = linearize(traceable, *primals)
+  else:
+    out_primal, pval, jaxpr, consts, aux = linearize(traceable, *primals, has_aux=True)
   def vjp_(ct):
     ct = ignore_consts(ct, pval)
     dummy_primal_and_ct = pack((core.unit, ct))
-    _, arg_cts = backward_pass(jaxpr, consts, (), dummy_primal_and_ct)
+    dummy_args = (None,) * len(jaxpr.invars)
+    _, arg_cts = backward_pass(jaxpr, consts, (), dummy_args, dummy_primal_and_ct)
     return instantiate_zeros(pack(primals), arg_cts[1])
 
-  return out_primal, vjp_
+  if not has_aux:
+    return out_primal, vjp_
+  else:
+    return out_primal, vjp_, aux
 
 def ignore_consts(ct, pval):
   aval, const = pval
@@ -100,7 +137,7 @@ def unpair_pval(pval):
     aval_1, aval_2 = aval
     return (aval_1, const_1), (aval_2, const_2)
 
-def backward_pass(jaxpr, consts, freevar_vals, cotangent_in):
+def backward_pass(jaxpr, consts, freevar_vals, args, cotangent_in):
   def write_cotangent(v, ct):
     # assert v not in primal_env
     if ct is not None:
@@ -109,21 +146,35 @@ def backward_pass(jaxpr, consts, freevar_vals, cotangent_in):
   def read_cotangent(v):
     return ct_env.get(v, zero)
 
-  primal_env = {v: val
-                for v, val in zip(jaxpr.freevars, freevar_vals)
-                if val is not None}
-  primal_env.update(zip(jaxpr.constvars, consts))
-  ct_env = {jaxpr.outvar: cotangent_in}
+  def read_primal(v):
+    if type(v) is Literal:
+      return v.val
+    else:
+      return primal_env.get(v)
 
+  def write_primal(v, val):
+    if val is not None:
+      primal_env[v] = val
+
+  primal_env = {}
+  core.pat_fmap(write_primal, jaxpr.constvars, consts)
+  core.pat_fmap(write_primal, jaxpr.freevars, freevar_vals)
+  core.pat_fmap(write_primal, jaxpr.invars, args)
+
+  ct_env = {jaxpr.outvar: cotangent_in}
   for eqn in jaxpr.eqns[::-1]:
     cts_in = map(read_cotangent, eqn.outvars)
     ct_in = TangentTuple(cts_in) if eqn.destructure else cts_in[0]
-    invals = map(primal_env.get, eqn.invars)
+    if not eqn.restructure:
+      invals = map(read_primal, eqn.invars)
+    else:
+      invals = [tuple(map(read_primal, v)) if type(v) is tuple
+                else read_primal(v) for v in eqn.invars]
     if eqn.bound_subjaxprs:
       subjaxprs, sub_consts, sub_freevar_vals = unzip3([
           (subjaxpr,
-           map(primal_env.get, const_vars),
-           map(primal_env.get, bound_vars))
+           map(read_primal, const_vars),
+           map(read_primal, bound_vars))
           for subjaxpr, const_vars, bound_vars in eqn.bound_subjaxprs])
       cts_out, ct_free_vars_out = get_primitive_transpose(eqn.primitive)(
           eqn.params, subjaxprs, sub_consts, sub_freevar_vals, invals, ct_in)
@@ -136,14 +187,35 @@ def backward_pass(jaxpr, consts, freevar_vals, cotangent_in):
 
     if cts_out is zero:
       cts_out = [zero for _ in eqn.invars]
+    if not eqn.restructure:
+      map(write_cotangent, eqn.invars, cts_out)
+    else:
+      [map(write_cotangent, v, ct) if type(v) is tuple
+       else write_cotangent(v, ct) for v, ct in zip(eqn.invars, cts_out)]
 
-    for var, ct in zip(eqn.invars, cts_out):
-      write_cotangent(var, ct)
-
-  cotangents_out = map(read_cotangent, jaxpr.invars)
-  freevar_cts = map(read_cotangent, jaxpr.freevars)
+  freevar_cts = core.pat_fmap(read_cotangent, jaxpr.freevars)
+  cotangents_out = core.pat_fmap(lambda v, _: read_cotangent(v), jaxpr.invars, None)
+  cotangents_out = tuple(map(pack_cotangents_like_caller, args, cotangents_out))
   return freevar_cts, cotangents_out
 
+def pack_cotangents_like_caller(arg, ct):
+  if type(arg) is tuple:
+    return tuple(map(pack_cotangents_like_caller, arg, ct))
+  elif arg is None:
+    return recursively_pack(ct)
+  else:
+    return None
+
+def recursively_pack(ct):
+  if type(ct) is tuple:
+    ct = tuple(map(recursively_pack, ct))
+    if any(elt is zero or isinstance(elt, TangentTuple) for elt in ct):
+      return TangentTuple(ct)
+    else:
+      return pack(ct)
+  else:
+    return ct
+
 def get_primitive_transpose(p):
   try:
     return primitive_transposes[p]
@@ -185,16 +257,11 @@ class JVPTrace(Trace):
     tangents = [t.tangent for t in tracers]
     nonzero_tangents, in_tree_def = tree_to_jaxtuples(tangents)
     f, out_tree_def = traceable(jvp_subtrace(f, self.master), in_tree_def)
-    if call_primitive is pxla.xla_pcall_p:
-      in_ax, out_ax = params['in_axes'], params['out_axes']
-      new_params = dict(params, in_axes=(in_ax, in_ax), out_axes=(out_ax, out_ax))
-      result = call_primitive.bind(f, pack(primals), nonzero_tangents, **new_params)
-    else:
-      result = call_primitive.bind(f, pack(primals), nonzero_tangents, **params)
+    result = call_primitive.bind(f, pack(primals), nonzero_tangents, **params)
     primal_out, tangent_out = build_tree(out_tree_def(), result)
     return JVPTracer(self, primal_out, tangent_out)
 
-  def post_process_call(self, _, out_tracer):
+  def post_process_call(self, call_primitive, out_tracer, params):
     out_jtuple, tree_def = tree_to_jaxtuples((out_tracer.primal, out_tracer.tangent))
     master = self.master
     def todo(x):
@@ -242,11 +309,11 @@ class JVPTracer(Tracer):
   __slots__ = ['primal', 'tangent']
 
   def __init__(self, trace, primal, tangent):
+    if not core.skip_checks:
+      _primal_tangent_shapes_match(primal, tangent)
     self.trace = trace
     self.primal = primal
     self.tangent = tangent
-    # TODO(mattjj,dougalm): behind skip_checks, check primal/tangent shapes and
-    # dtypes agree (up to jax_enable_x64 flag)
 
   @property
   def aval(self):
@@ -265,6 +332,15 @@ class JVPTracer(Tracer):
     else:
       return self
 
+def _primal_tangent_shapes_match(primal, tangent):
+  if type(tangent) is TangentTuple:
+    for p, t in zip(primal, tangent):
+      _primal_tangent_shapes_match(p, t)
+  elif tangent is not zero:
+    primal_aval = raise_to_shaped(get_aval(primal))
+    tangent_aval = raise_to_shaped(get_aval(tangent))
+    assert primal_aval == tangent_aval
+
 # -------------------- Primitives --------------------
 
 
@@ -278,7 +354,6 @@ def deflinear(primitive, transpose_rule):
   primitive_jvps[primitive] = partial(linear_jvp, primitive)
   primitive_transposes[primitive] = partial(linear_transpose, transpose_rule)
 
-
 def linear_jvp(primitive, primals, tangents, **params):
   val_out = primitive.bind(*primals, **params)
   if all(tangent is zero for tangent in tangents):
@@ -287,7 +362,6 @@ def linear_jvp(primitive, primals, tangents, **params):
     tangents = map(instantiate_zeros, primals, tangents)
     return val_out, primitive.bind(*tangents, **params)
 
-
 def linear_transpose(transpose_rule, cotangent, *args, **kwargs):
   return zero if cotangent is zero else transpose_rule(cotangent, **kwargs)
 
@@ -299,23 +373,20 @@ def defjvp(primitive, *jvprules):
 
 def standard_jvp(jvprules, primitive, primals, tangents, **params):
   val_out = primitive.bind(*primals, **params)
-  tangents_out = (rule(t, *primals, **params) for rule, t in zip(jvprules, tangents)
-                  if rule is not None and t is not zero)
+  tangents_out = [rule(t, *primals, **params) for rule, t in zip(jvprules, tangents)
+                  if rule is not None and t is not zero]
   return val_out, reduce(add_tangents, tangents_out, zero)
 
-
 def defjvp2(primitive, *jvprules):
   assert isinstance(primitive, Primitive)
   primitive_jvps[primitive] = partial(standard_jvp2, jvprules, primitive)
 
-
 def standard_jvp2(jvprules, primitive, primals, tangents, **params):
   val_out = primitive.bind(*primals, **params)
   tangents_out = (rule(t, val_out, *primals, **params) for rule, t in zip(jvprules, tangents)
                   if rule is not None and t is not zero)
   return val_out, reduce(add_tangents, tangents_out, zero)
 
-
 def add_tangents(x, y):
   if x is zero:
     return y
@@ -325,6 +396,57 @@ def add_tangents(x, y):
     return add_jaxvals(x, y)
 
 
+def defvjp_all(prim, custom_vjp):
+  name = prim.name
+
+  def fun_jvp(xs, ts):
+    ts = map(instantiate_zeros, xs, ts)  # TODO(mattjj): avoid instantiation?
+    primal_out, tangent_out = fun_jvp_p.bind(pack(xs), pack(ts))
+    return primal_out, tangent_out
+  primitive_jvps[prim] = fun_jvp
+
+  fun_jvp_p = core.Primitive('{name}_jvp'.format(name=name))
+  def fun_jvp_partial_eval(trace, *tracers):
+    primals_tracer, tangents_tracer = tracers
+    primal_out, vjp_py = custom_vjp(*primals_tracer)
+
+    in_aval = raise_to_shaped(get_aval(primal_out))
+    ct_pval = pe.PartialVal((in_aval, core.unit))
+    vjp_jaxpr, out_pval, residuals = pe.trace_unwrapped_to_jaxpr(
+        lambda ct: pack(vjp_py(ct)), (ct_pval,))
+    out_pv, out_const = out_pval
+    tangent_out = fun_lin_p.bind(out_const, pack(residuals), tangents_tracer,
+                                 in_aval=in_aval, out_pv=out_pv, vjp_jaxpr=vjp_jaxpr)
+
+    return pack((primal_out, tangent_out))
+  pe.custom_partial_eval_rules[fun_jvp_p] = fun_jvp_partial_eval
+
+  fun_lin_p = core.Primitive('{name}_lin'.format(name=name))
+  fun_lin_p.def_abstract_eval(lambda c, r, ts, in_aval, out_pv, vjp_jaxpr: in_aval)
+  def fun_lin_transpose(ct, out_const, residuals, ts, in_aval, out_pv, vjp_jaxpr):
+    assert ts is None and out_const is not None and residuals is not None
+    ans = core.eval_jaxpr(vjp_jaxpr, residuals, (), ct)
+    out = pe.merge_pvals(ans, pe.PartialVal((out_pv, out_const)))
+    return [None, None, out]
+  primitive_transposes[fun_lin_p] = fun_lin_transpose
+
+def defvjp(prim, *vjps):
+  def vjpmaker(*primals):
+    ans = prim.bind(*primals)
+    vjpfun = lambda ct: [vjp(ct, *primals) if vjp else zeros_like_jaxval(x)
+                         for x, vjp in zip(primals, vjps)]
+    return ans, vjpfun
+  defvjp_all(prim, vjpmaker)
+
+def defvjp2(prim, *vjps):
+  def vjpmaker(*primals):
+    ans = prim.bind(*primals)
+    vjpfun = lambda ct: [vjp(ct, ans, *primals) if vjp else zeros_like_jaxval(x)
+                         for x, vjp in zip(primals, vjps)]
+    return ans, vjpfun
+  defvjp_all(prim, vjpmaker)
+
+
 def defbilinear_broadcasting(bcast, prim, lhs_rule, rhs_rule):
   assert isinstance(prim, Primitive)
   lhs_jvp = lambda g, x, y, **kwargs: prim.bind(bcast(g, y), y, **kwargs)
@@ -333,7 +455,6 @@ def defbilinear_broadcasting(bcast, prim, lhs_rule, rhs_rule):
   primitive_transposes[prim] = partial(bilinear_transpose, lhs_rule, rhs_rule)
 defbilinear = partial(defbilinear_broadcasting, lambda g, x: g)
 
-
 def bilinear_transpose(lhs_rule, rhs_rule, cotangent, x, y, **kwargs):
   assert (x is None) ^ (y is None)
   if x is None:
@@ -348,7 +469,6 @@ def defjvp_zero(primitive):
   assert isinstance(primitive, Primitive)
   primitive_jvps[primitive] = partial(zero_jvp, primitive)
 
-
 def zero_jvp(primitive, primals, tangents, **params):
   return primitive.bind(*primals, **params), zero
 
@@ -359,6 +479,19 @@ deflinear(core.pack_p, lambda t: list(t) if t is not zero else zero)
 deflinear(add_jaxvals_p, lambda t: (t, t))
 
 
+def instantiate_zeros_at(instantiate, example, tangent):
+  t = type(instantiate)
+  if t is tuple:
+    # note to future selves: it wasn't clear whether to pack here
+    return TangentTuple(map(instantiate_zeros_at, instantiate, example, tangent))
+  elif t is bool:
+    if instantiate:
+      return instantiate_zeros(example, tangent)
+    else:
+      return tangent
+  else:
+    raise TypeError(t)
+
 def instantiate_zeros(example, tangent):
   if tangent is zero:
     return zeros_like_jaxval(example)
@@ -367,18 +500,26 @@ def instantiate_zeros(example, tangent):
   else:
     return tangent
 
+def instantiate_zeros_aval(aval, tangent):
+  if tangent is zero:
+    return zeros_like_aval(aval)
+  elif isinstance(tangent, TangentTuple):
+    return pack(map(instantiate_zeros_aval, aval, tangent))
+  else:
+    return tangent
+
 @transformation_with_aux
 def traceable(in_tree_def, new_primals, new_tangents):
   new_tangents = build_tree(in_tree_def, new_tangents)
-  primal_out, tangent_out = yield new_primals, new_tangents
+  primal_out, tangent_out = yield (new_primals, new_tangents), {}
   out_jtuple, tree_def = tree_to_jaxtuples((primal_out, tangent_out))
   yield out_jtuple, tree_def
 
 @transformation_with_aux
 def transposed_fun(jaxpr, in_tree_def, args):
-  consts, freevar_vals, ct = args
-  ct, freevar_vals = build_tree(in_tree_def, (ct, freevar_vals))
-  freevar_cts, cotangents_out = yield jaxpr, consts, freevar_vals, ct
+  args, consts, freevar_vals, ct = args
+  args, ct, freevar_vals = build_tree(in_tree_def, (args, ct, freevar_vals))
+  freevar_cts, cotangents_out = yield (jaxpr, consts, freevar_vals, args, ct), {}
   out_jtuple, tree_def = tree_to_jaxtuples((cotangents_out, freevar_cts))
   yield out_jtuple, tree_def
 
@@ -387,26 +528,96 @@ def call_transpose(primitive, params, jaxpr, consts, freevar_vals, args, ct):
   consts, = consts
   freevar_vals, = freevar_vals
   assert isinstance(jaxpr, core.Jaxpr)
-  assert all(a is None for a in args), "TODO(dougalm): handle non-tangent primal args"
-  (ct, freevar_vals), in_tree_def = tree_to_jaxtuples((ct, freevar_vals))
+  (args, ct, freevar_vals), in_tree_def = tree_to_jaxtuples((args, ct, freevar_vals))
   fun = wrap_init(backward_pass)
   fun, out_tree_def = transposed_fun(fun, jaxpr, in_tree_def)
-  all_args = pack((pack(consts), pack(freevar_vals), ct))
+  all_args = pack((pack(args), pack(consts), pack(freevar_vals), ct))
   # TODO(dougalm): consider signalling to bind that no traces in fun closure
-  if primitive is pxla.xla_pcall_p:
-    in_axes, out_axes = params['in_axes'], params['out_axes']
-    trans_in_axes = (None, None, out_axes),
-    trans_out_axes = (in_axes, None)
-    new_params = dict(params, in_axes=trans_in_axes, out_axes=trans_out_axes)
-    ans = primitive.bind(fun, all_args, **new_params)
-  else:
-    ans = primitive.bind(fun, all_args, **params)
+  ans = primitive.bind(fun, all_args, **params)
   return build_tree(out_tree_def(), ans)
 
+@transformation_with_aux
+def transposed_mapped(jaxpr, in_tree_def, freevar_vals, args):
+  args, consts, ct = args
+  args, ct = build_tree(in_tree_def, (args, ct))
+  freevar_cts, cotangents_out = yield (jaxpr, consts, freevar_vals, args, ct), {}
+  out_jtuple, tree_def = tree_to_jaxtuples((cotangents_out, freevar_cts))
+  yield out_jtuple, tree_def
+
+def map_transpose(primitive, params, jaxpr, consts, freevar_vals, args, ct):
+  jaxpr, = jaxpr
+  consts, = consts
+  freevar_vals, = freevar_vals
+  (args, ct), in_tree_def = tree_to_jaxtuples((args, ct))
+  fun = wrap_init(backward_pass)
+  fun, out_tree_def = transposed_mapped(fun, jaxpr, in_tree_def, tuple(freevar_vals))
+  all_args = pack((pack(args), pack(consts), ct))
+  ans = primitive.bind(fun, all_args, **params)
+  cts_out, freevar_cts = build_tree(out_tree_def(), ans)
+  freevar_cts = tree_map(lambda x: x.sum(0), freevar_cts)
+  return cts_out, freevar_cts
+
+def jaxpr_as_fun(jaxpr, consts, *args):
+  return core.eval_jaxpr(jaxpr, consts, (), *args)
+
+def get_nonzeros(tangent):
+  if tangent is zero:
+    return False
+  elif isinstance(tangent, TangentTuple):
+    return tuple(map(get_nonzeros, tangent))
+  else:
+    return True
+
+def put_zeros(pack, isnonzero, x):
+  if isnonzero is True:
+    return x
+  elif isnonzero is False:
+    return zero
+  else:
+    return pack(map(partial(put_zeros, pack), isnonzero, x))
+
+def strip_zeros(unit, pack, isnonzero, x):
+  if isnonzero is True:
+    return x
+  elif isnonzero is False:
+    return unit
+  else:
+    return pack(map(partial(strip_zeros, unit, pack), isnonzero, x))
+
+@transformation_with_aux
+def f_jvp_traceable(nonzero_components, *primal_tangent_pairs):
+  primals, tangents = unzip2(primal_tangent_pairs)
+  tangents_zeros = map(partial(put_zeros, TangentTuple), nonzero_components, tangents)
+  primal_out, tangent_out = yield (primals, tangents_zeros), {}
+  # TODO check output is tuple
+  nonzeros_out = get_nonzeros(tangent_out)
+  tangent_out_nonzero = strip_zeros(core.unit, pack, nonzeros_out, tangent_out)
+  primal_tangent_pairs_out = [pack((p, t)) for p, t in zip(primal_out, tangent_out_nonzero)]
+  yield pack(primal_tangent_pairs_out), nonzeros_out
+
+def jvp_jaxpr(jaxpr, nonzeros, instantiate):
+  # jaxpr :: d -> a -> b -> (c1, c2)
+  # avals = (d, a, b)
+  # f :: d -> a -> b -> (c1, c2)
+  f = wrap_init(partial(jaxpr_as_fun, jaxpr.jaxpr, jaxpr.literals))
+  f_jvp, out_nonzeros = f_jvp_traceable(jvp(f, instantiate=instantiate), nonzeros)
+  # f_jvp :: (d, d') -> (a, a') -> (b, b') -> ((c1, c1'), (c2, c2'))
+  tangent_avals = map(partial(strip_zeros, core.AbstractTuple(()), core.AbstractTuple),
+                      nonzeros, jaxpr.in_avals)
+  pt_pvals = [pe.PartialVal((core.AbstractTuple((p_aval, t_aval)), core.unit))
+              for p_aval, t_aval in zip(jaxpr.in_avals, tangent_avals)]
+  jaxpr_out, pval_out, literals_out = pe.trace_to_jaxpr(
+      f_jvp, pt_pvals, instantiate=True)
+  # jaxpr_out :: (d, d') -> (a, a') -> (b, b') -> ((c1, c1'), (c2, c2'))
+  # out_nonzeros :: (nonzeros(c1), nonzeros(c2))
+  in_avals = tuple(map(core.AbstractTuple, zip(jaxpr.in_avals, tangent_avals)))
+  out_aval, _ = pval_out
+  jaxpr_out = core.TypedJaxpr(jaxpr_out, literals_out, in_avals, out_aval)
+  return jaxpr_out, out_nonzeros()
+
+
 primitive_transposes[core.call_p] = partial(call_transpose, call_p)
 primitive_transposes[pe.compiled_call_p] = partial(call_transpose, pe.compiled_call_p)
-primitive_transposes[xla.xla_call_p] = partial(call_transpose, xla.xla_call_p)
-primitive_transposes[pxla.xla_pcall_p] = partial(call_transpose, pxla.xla_pcall_p)
 
 
 tree_to_jaxtuples = partial(process_pytree, pack)
