@@ -26,43 +26,43 @@ from six.moves import reduce
 
 from .. import core
 from ..core import Trace, Tracer, new_master, pack, AbstractTuple, JaxTuple
-from ..abstract_arrays import ShapedArray, make_shaped_array, array_types
+from ..abstract_arrays import ShapedArray, make_shaped_array, array_types, raise_to_shaped
 from ..ad_util import add_jaxvals_p, zeros_like_p, zeros_like_jaxval
 from ..linear_util import transformation, transformation_with_aux, wrap_init
 from ..tree_util import register_pytree_node
 from ..util import unzip2, partial, safe_map
+from . import xla
 
 map = safe_map
 
 
-def batch(fun, in_vals, in_dims, out_dim_target):
+def batch(fun, in_vals, in_dims, out_dim_dst):
   sizes = reduce(set.union, map(dimsize, in_dims, in_vals))
   if not sizes:
     return fun.call_wrapped(*in_vals), None  # no mapped dimensions
   elif len(sizes) == 1:
-    out_val, out_dim = batch_transform(fun).call_wrapped(in_vals, in_dims)
-    return moveaxis(sizes.pop(), out_dim_target, out_dim, out_val)
+    sz = sizes.pop()
+    return batch_transform(fun, sz, in_dims, out_dim_dst).call_wrapped(in_vals)
   else:
     raise TypeError("got inconsistent map dimension sizes: {}".format(sizes))
 
 
-# TODO(mattjj,dougalm): could call batch_subtrace here (a bit redundant)
 @transformation
-def batch_transform(vals, dims):
+def batch_transform(size, in_dims, out_dim_dst, vals):
   with new_master(BatchTrace) as master:
     trace = BatchTrace(master, core.cur_sublevel())
-    in_tracers = map(partial(BatchTracer, trace), vals, dims)
-    out_tracer = yield in_tracers
+    in_tracers = map(partial(BatchTracer, trace), vals, in_dims)
+    out_tracer = yield in_tracers, {}
     out_tracer = trace.full_raise(out_tracer)
     out_val, out_dim = out_tracer.val, out_tracer.batch_dim
     del master
-  yield (out_val, out_dim)
+  yield moveaxis(size, out_dim_dst, out_dim, out_val)
 
 
 @transformation_with_aux
 def batch_subtrace(master, dims, *vals):
   trace = BatchTrace(master, core.cur_sublevel())
-  ans = yield map(partial(BatchTracer, trace), vals, dims)
+  ans = yield map(partial(BatchTracer, trace), vals, dims), {}
   out_tracer = trace.full_raise(ans)
   out_val, out_dim = out_tracer.val, out_tracer.batch_dim
   yield out_val, out_dim
@@ -131,12 +131,12 @@ class BatchTrace(Trace):
       val_out = call_primitive.bind(f, *vals, **params)
       return BatchTracer(self, val_out, dim_out())
 
-  def post_process_call(self, _, out_tracer):
+  def post_process_call(self, call_primitive, out_tracer, params):
     val, dim = out_tracer.val, out_tracer.batch_dim
     master = self.master
     def todo(x):
       trace = BatchTrace(master, core.cur_sublevel())
-      return BatchTracer(trace, val, dim)
+      return BatchTracer(trace, x, dim)
 
     return val, todo
 
@@ -161,14 +161,6 @@ def shaped_aval(x):
   except KeyError:
     raise TypeError("{} is not a valid type for batching".format(type(x)))
 
-def raise_to_shaped(aval):
-  if type(aval) is AbstractTuple:
-    return AbstractTuple(map(raise_to_shaped, aval))
-  elif isinstance(aval, ShapedArray):
-    return ShapedArray(aval.shape, aval.dtype)
-  else:
-    raise TypeError(type(aval))
-
 def remove_batch_dim_from_aval(bdim, aval):
   t = type(aval)
   if t is AbstractTuple:
@@ -186,14 +178,32 @@ def remove_batch_dim_from_aval(bdim, aval):
   else:
     raise TypeError(t)
 
+def add_batch_dim_to_aval(bdim, size, aval):
+  t = type(aval)
+  if t is AbstractTuple:
+    if type(bdim) is tuple:
+      return AbstractTuple(map(add_batch_dim_to_aval, bdim, size, aval))
+    else:
+      return AbstractTuple(map(partial(add_batch_dim_to_aval, bdim, size), aval))
+  elif t is ShapedArray:
+    if bdim is None:
+      return ShapedArray(aval.shape, aval.dtype)
+    else:
+      assert 0 <= bdim <= aval.ndim
+      batched_shape = tuple(onp.insert(aval.shape, bdim, size))
+      return ShapedArray(batched_shape, aval.dtype)
+  else:
+    raise TypeError(t)
+
 pytype_aval_mappings = {}
 
 def shaped_jaxtuple(xs):
   return AbstractTuple(map(shaped_aval, xs))
 
 pytype_aval_mappings[JaxTuple] = shaped_jaxtuple
+pytype_aval_mappings[xla.DeviceTuple] = xla.pytype_aval_mappings[xla.DeviceTuple]
 
-for t in array_types:
+for t in it.chain(array_types, [xla.DeviceArray]):
   pytype_aval_mappings[t] = make_shaped_array
 
 
@@ -241,11 +251,13 @@ def reducer_batcher(prim, batched_args, batch_dims, axes, **params):
 
 def add_batched(batched_args, batch_dims):
   bdx, bdy = batch_dims
+  xs, ys = batched_args
   if bdx == bdy:
-    xs, ys = batched_args
     return add_jaxvals_p.bind(xs, ys), bdx
   else:
-    xs, ys = map(bdim_at_front, batched_args, batch_dims)
+    sz = (dimsize(bdx, xs) | dimsize(bdy, ys)).pop()
+    move_bdim = partial(bdim_at_front, broadcast_size=sz, force_broadcast=True)
+    xs, ys = map(move_bdim, batched_args, batch_dims)
     return add_jaxvals_p.bind(xs, ys), 0
 primitive_batchers[add_jaxvals_p] = add_batched
 
@@ -267,55 +279,55 @@ primitive_batchers[zeros_like_p] = zeros_like_batched
 
 
 def bdim_at_front(x, bdim, broadcast_size=1, force_broadcast=False):
-  if bdim is None:
-    return broadcast(x, broadcast_size, force_broadcast=force_broadcast)
-  else:
-    return move_dim_to_front(x, bdim)
+  return moveaxis(broadcast_size, 0, bdim, x, force_broadcast=force_broadcast)
 
 def move_dim_to_front(x, dim):
-  aval = get_aval(x)
-  if type(aval) is AbstractTuple:
-    return pack(map(partial(move_dim_to_front, dim=dim), x))
-  elif isinstance(aval, ShapedArray):
-    assert 0 <= dim < onp.ndim(x)
-    if dim == 0:
-      return x
-    else:
-      perm = (dim,) + tuple(range(dim)) + tuple(range(dim + 1, onp.ndim(x)))
-      return x.transpose(perm)
-  else:
-    raise TypeError(type(x))
+  assert dim is not None
+  return moveaxis(None, 0, dim, x)
 
 def dimsize(dim, x):
-  aval = get_aval(x)
-  if type(aval) is AbstractTuple:
-    return reduce(set.union, map(partial(dimsize, dim), x))
-  elif type(dim) is int:
-    return {x.shape[dim]}
-  elif dim is None:
-    return set()
-  else:
-    raise TypeError(type(dim))
+  return _dimsize(dim, get_aval(x), x)
 
-def moveaxis(sz, dst, src, x):
-  aval = get_aval(x)
+def _dimsize(dim, aval, x):
+  if type(aval) is AbstractTuple:
+    if type(dim) is tuple:
+      return reduce(set.union, map(_dimsize, dim, aval, x))
+    elif type(dim) is int:
+      return reduce(set.union, map(partial(_dimsize, dim), aval, x))
+    elif dim is None:
+      return set()
+    else:
+      raise TypeError(type(dim))
+  else:
+    if type(dim) is int:
+      return {x.shape[dim]}
+    elif dim is None:
+      return set()
+    else:
+      raise TypeError(type(dim))
+
+def moveaxis(sz, dst, src, x, force_broadcast=True):
+  return _moveaxis(sz, dst, src, get_aval(x), x, force_broadcast)
+
+# TODO(mattjj): not passing force_broadcast recursively... intentional?
+def _moveaxis(sz, dst, src, aval, x, force_broadcast=True):
   if type(aval) is AbstractTuple:
     if type(src) is tuple and type(dst) is tuple:
-      return pack(map(partial(moveaxis, sz), dst, src, x))
+      return pack(map(partial(_moveaxis, sz), dst, src, aval, x))
     elif type(src) is tuple:
-      return pack(map(partial(moveaxis, sz, dst), src, x))
+      return pack(map(partial(_moveaxis, sz, dst), src, aval, x))
     elif type(dst) is tuple:
       srcs = (src,) * len(dst)
-      return pack(map(partial(moveaxis, sz), dst, srcs, x))
+      return pack(map(partial(_moveaxis, sz), dst, srcs, aval, x))
     else:
-      return pack(map(partial(moveaxis, sz, dst, src), x))
+      return pack(map(partial(_moveaxis, sz, dst, src), aval, x))
   elif isinstance(aval, ShapedArray):
     dst_ = (dst % aval.ndim) if dst is not None and aval.ndim else dst
     if src == dst_:
       return x
     else:
       if src is None:
-        x = broadcast(x, sz, force_broadcast=True)
+        x = broadcast(x, sz, force_broadcast=force_broadcast)
         src = 0
         dst_ = dst % (aval.ndim + 1)
       if src == dst_:
@@ -328,9 +340,12 @@ def moveaxis(sz, dst, src, x):
     raise TypeError(type(aval))
 
 def broadcast(x, sz, force_broadcast=False):
-  aval = get_aval(x)
+  return _broadcast(sz, get_aval(x), x, force_broadcast)
+
+# TODO(mattjj): not passing force_broadcast recursively... intentional?
+def _broadcast(sz, aval, x, force_broadcast=False):
   if type(aval) is AbstractTuple:
-    return pack(map(partial(broadcast, sz=sz), x))
+    return pack(map(partial(_broadcast, sz), aval, x))
   elif isinstance(aval, ShapedArray):
     # for scalars, maybe don't actually broadcast
     if not onp.ndim(x) and not force_broadcast:
