@@ -37,11 +37,21 @@ from jaxlib import lapack
 def cholesky(x, symmetrize_input=True):
   if symmetrize_input:
     x = symmetrize(x)
-  return cholesky_p.bind(x)
+  return np.tril(cholesky_p.bind(x))
 
-def eigh(x, lower=True): return eigh_p.bind(x, lower=lower)
+def eig(x):
+  w, vl, vr = eig_p.bind(x)
+  return w, vl, vr
 
-def lu(x): return lu_p.bind(x)
+def eigh(x, lower=True, symmetrize_input=True):
+  if symmetrize_input:
+    x = symmetrize(x)
+  v, w = eigh_p.bind(x, lower=lower)
+  return v, w
+
+def lu(x):
+  lu, pivots = lu_p.bind(x)
+  return lu, pivots
 
 def qr(x, full_matrices=True):
   q, r = qr_p.bind(x, full_matrices=full_matrices)
@@ -77,7 +87,7 @@ _cpu_lapack_types = {np.float32, np.float64, np.complex64, np.complex128}
 def cholesky_jvp_rule(primals, tangents):
   x, = primals
   sigma_dot, = tangents
-  L = cholesky_p.bind(x)
+  L = np.tril(cholesky_p.bind(x))
 
   # Forward-mode rule from https://arxiv.org/pdf/1602.07527.pdf
   phi = lambda X: np.tril(X) / (1 + np.eye(X.shape[-1], dtype=X.dtype))
@@ -109,7 +119,48 @@ def cholesky_cpu_translation_rule(c, operand):
     # TODO(phawkins): support LAPACK primitives in batched mode.
     return c.Cholesky(operand)
 
-xla.backend_specific_translations['Host'][cholesky_p] = cholesky_cpu_translation_rule
+xla.backend_specific_translations['cpu'][cholesky_p] = cholesky_cpu_translation_rule
+
+# Asymmetric eigendecomposition
+
+def eig_impl(operand):
+  return xla.apply_primitive(eig_p, operand)
+
+def eig_translation_rule(c, operand):
+  raise NotImplementedError(
+    "Nonsymmetric eigendecomposition is only implemented on the CPU backend")
+
+def eig_abstract_eval(operand):
+  if isinstance(operand, ShapedArray):
+    if operand.ndim < 2 or operand.shape[-2] != operand.shape[-1]:
+      raise ValueError("Argument to nonsymmetric eigendecomposition must have "
+                       "shape [..., n, n], got shape {}".format(operand.shape))
+
+    batch_dims = operand.shape[:-2]
+    n = operand.shape[-1]
+    vl = vr = ShapedArray(batch_dims + (n, n), operand.dtype)
+    w = ShapedArray(batch_dims + (n,), lax.lax._complex_basetype(operand.dtype))
+  else:
+    w = vl = vr = operand
+  return core.AbstractTuple((w, vl, vr))
+
+def eig_cpu_translation_rule(c, operand):
+  out = lapack.jax_geev(c, operand)
+  return c.Tuple(c.GetTupleElement(out, 0), c.GetTupleElement(out, 1),
+                 c.GetTupleElement(out, 2))
+
+def eig_batching_rule(batched_args, batch_dims):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.bdim_at_front(x, bd)
+  return eig_p.bind(x), 0
+
+eig_p = Primitive('eig')
+eig_p.def_impl(eig_impl)
+eig_p.def_abstract_eval(eig_abstract_eval)
+xla.translations[eig_p] = eig_translation_rule
+xla.backend_specific_translations['cpu'][eig_p] = eig_cpu_translation_rule
+batching.primitive_batchers[eig_p] = eig_batching_rule
 
 
 # Symmetric/Hermitian eigendecomposition
@@ -126,31 +177,58 @@ def eigh_abstract_eval(operand, lower):
   if isinstance(operand, ShapedArray):
     if operand.ndim < 2 or operand.shape[-2] != operand.shape[-1]:
       raise ValueError(
-        "Argument to symmetric eigendecomposition must have shape [..., n, n]")
+        "Argument to symmetric eigendecomposition must have shape [..., n, n],"
+        "got shape {}".format(operand.shape))
 
     batch_dims = operand.shape[:-2]
     n = operand.shape[-1]
     v = ShapedArray(batch_dims + (n, n), operand.dtype)
-    w = ShapedArray(batch_dims + (n,), operand.dtype)
+    w = ShapedArray(batch_dims + (n,), lax.lax._complex_basetype(operand.dtype))
   else:
     v, w = operand, operand
   return core.AbstractTuple((v, w))
 
 def eigh_cpu_translation_rule(c, operand, lower):
-  shape = c.GetShape(operand)
-  dtype = shape.element_type().type
-  if len(shape.dimensions()) == 2 and dtype in _cpu_lapack_types:
-    out = lapack.jax_syevd(c, operand, lower=lower)
-    return c.Tuple(c.GetTupleElement(out, 0), c.GetTupleElement(out, 1))
-  else:
-    raise NotImplementedError(
-        "Only unbatched eigendecomposition is implemented on CPU")
+  out = lapack.jax_syevd(c, operand, lower=lower)
+  return c.Tuple(c.GetTupleElement(out, 0), c.GetTupleElement(out, 1))
+
+def eigh_jvp_rule(primals, tangents, lower):
+  # Derivative for eigh in the simplest case of distinct eigenvalues.
+  # This is classic nondegenerate perurbation theory, but also see
+  # https://people.maths.ox.ac.uk/gilesm/files/NA-08-01.pdf
+  # The general solution treating the case of degenerate eigenvalues is
+  # considerably more complicated. Ambitious readers may refer to the general
+  # methods below or refer to degenerate perturbation theory in physics.
+  # https://www.win.tue.nl/analysis/reports/rana06-33.pdf and
+  # https://people.orie.cornell.edu/aslewis/publications/99-clarke.pdf
+  a, = primals
+  a_dot, = tangents
+  v, w = eigh_p.bind(symmetrize(a), lower=lower)
+  # for complex numbers we need eigenvalues to be full dtype of v, a:
+  w = w.astype(a.dtype)
+  eye_n = np.eye(a.shape[-1], dtype=a.dtype)
+  # carefully build reciprocal delta-eigenvalue matrix, avoiding NaNs.
+  Fmat = np.reciprocal(eye_n + w - w[..., np.newaxis]) - eye_n
+  # eigh impl doesn't support batch dims, but future-proof the grad.
+  dot = lax.dot if a.ndim == 2 else lax.batch_matmul
+  vdag_adot_v = dot(dot(_H(v), a_dot), v)
+  dv = dot(v, np.multiply(Fmat, vdag_adot_v))
+  dw = np.diagonal(vdag_adot_v)
+  return core.pack((v, w)), core.pack((dv, dw))
+
+def eigh_batching_rule(batched_args, batch_dims, lower):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.bdim_at_front(x, bd)
+  return eigh_p.bind(x, lower=lower), 0
 
 eigh_p = Primitive('eigh')
 eigh_p.def_impl(eigh_impl)
 eigh_p.def_abstract_eval(eigh_abstract_eval)
 xla.translations[eigh_p] = eigh_translation_rule
-xla.backend_specific_translations['Host'][eigh_p] = eigh_cpu_translation_rule
+ad.primitive_jvps[eigh_p] = eigh_jvp_rule
+xla.backend_specific_translations['cpu'][eigh_p] = eigh_cpu_translation_rule
+batching.primitive_batchers[eigh_p] = eigh_batching_rule
 
 
 
@@ -189,11 +267,26 @@ def triangular_solve_jvp_rule_a(
 
 def triangular_solve_transpose_rule(
     cotangent, a, b, left_side, lower, transpose_a, conjugate_a):
+  # Triangular solve is linear in its first argument and nonlinear in its second
+  # argument, similar to `div`. We need both a JVP rule and a transpose rule
+  # for the first argument.
   assert a is not None and b is None
   cotangent_b = triangular_solve(a, cotangent, left_side, lower,
                                  not transpose_a, conjugate_a)
   return [None, cotangent_b]
 
+
+def triangular_solve_batching_rule(batched_args, batch_dims, left_side,
+                                   lower, transpose_a, conjugate_a):
+  x, y = batched_args
+  bx, by = batch_dims
+  size = next(t.shape[i] for t, i in zip(batched_args, batch_dims)
+              if i is not None)
+  x = batching.bdim_at_front(x, bx, size, force_broadcast=True)
+  y = batching.bdim_at_front(y, by, size, force_broadcast=True)
+  return triangular_solve(x, y, left_side=left_side, lower=lower,
+                          transpose_a=transpose_a, conjugate_a=conjugate_a), 0
+
 triangular_solve_p = standard_primitive(
     triangular_solve_shape_rule, triangular_solve_dtype_rule,
     'triangular_solve')
@@ -201,6 +294,7 @@ ad.defjvp2(triangular_solve_p,
            triangular_solve_jvp_rule_a,
            lambda g_b, _, a, b, **kws: triangular_solve(a, g_b, **kws))
 ad.primitive_transposes[triangular_solve_p] = triangular_solve_transpose_rule
+batching.primitive_batchers[triangular_solve_p] = triangular_solve_batching_rule
 
 
 def triangular_solve_cpu_translation_rule(
@@ -217,7 +311,7 @@ def triangular_solve_cpu_translation_rule(
     # TODO(phawkins): support BLAS primitives in batched mode.
     return c.TriangularSolve(a, b, left_side, lower, transpose_a, conjugate_a)
 
-xla.backend_specific_translations['Host'][triangular_solve_p] = triangular_solve_cpu_translation_rule
+xla.backend_specific_translations['cpu'][triangular_solve_p] = triangular_solve_cpu_translation_rule
 
 
 # LU decomposition
@@ -255,7 +349,7 @@ def lu_jvp_rule(primals, tangents):
 
   a_shape = np.shape(a)
   m, n = a_shape[-2:]
-  dtype = lax._dtype(a)
+  dtype = lax.dtype(a)
   k = min(m, n)
 
   permutation = lu_pivots_to_permutation(pivots, m)
@@ -294,29 +388,31 @@ def lu_jvp_rule(primals, tangents):
   return core.pack((lu, pivots)), ad.TangentTuple((lu_dot, ad_util.zero))
 
 
+def lu_batching_rule(batched_args, batch_dims):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.bdim_at_front(x, bd)
+  return lu_p.bind(x), 0
+
+
 lu_p = Primitive('lu')
 lu_p.def_impl(lu_impl)
 lu_p.def_abstract_eval(lu_abstract_eval)
 xla.translations[lu_p] = lu_translation_rule
 ad.primitive_jvps[lu_p] = lu_jvp_rule
+batching.primitive_batchers[lu_p] = lu_batching_rule
 
 def lu_cpu_translation_rule(c, operand):
   shape = c.GetShape(operand)
   dtype = shape.element_type().type
-  if len(shape.dimensions()) == 2 and dtype in _cpu_lapack_types:
-    out = lapack.jax_getrf(c, operand)
-    lu = c.GetTupleElement(out, 0)
-    # Subtract 1 from the pivot to get 0-based indices.
-    pivot = c.Sub(c.GetTupleElement(out, 1), c.ConstantS32Scalar(1))
-    # Throw away the `info` value, because we have no way to report errors.
-    return c.Tuple(lu, pivot)
-  else:
-    raise NotImplementedError("Only unbatched LU decomposition is implemented")
+  out = lapack.jax_getrf(c, operand)
+  lu = c.GetTupleElement(out, 0)
+  # Subtract 1 from the pivot to get 0-based indices.
+  pivot = c.Sub(c.GetTupleElement(out, 1), c.ConstantS32Scalar(1))
+  # Throw away the `info` value, because we have no way to report errors.
+  return c.Tuple(lu, pivot)
 
-# TODO(phawkins): The hasattr() test here is to avoid incompatibilities between
-# jax and an older jaxlib. Remove after a jaxlib release includes jax_getrf.
-if hasattr(lapack, "jax_getrf"):
-  xla.backend_specific_translations['Host'][lu_p] = lu_cpu_translation_rule
+xla.backend_specific_translations['cpu'][lu_p] = lu_cpu_translation_rule
 
 
 def lu_pivots_to_permutation(swaps, k):
@@ -332,8 +428,8 @@ def lu_pivots_to_permutation(swaps, k):
 
   n, = np.shape(swaps)
   permutation = np.arange(k)
-  _, permutation = lax.fori_loop(onp.array(0, onp.int32), onp.array(n, onp.int32),
-                                 body_fn, (swaps, permutation))
+  _, permutation = lax.fori_loop(
+      onp.array(0, onp.int32), onp.array(n, onp.int32), body_fn, (swaps, permutation))
   return permutation
 
 
@@ -377,11 +473,18 @@ def qr_jvp_rule(primals, tangents, full_matrices):
   dr = np.matmul(qt_dx_rinv - domega, r)
   return core.pack((q, r)), core.pack((dq, dr))
 
+def qr_batching_rule(batched_args, batch_dims, full_matrices):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.bdim_at_front(x, bd)
+  return qr_p.bind(x, full_matrices=full_matrices), 0
+
 qr_p = Primitive('qr')
 qr_p.def_impl(qr_impl)
 qr_p.def_abstract_eval(qr_abstract_eval)
 xla.translations[qr_p] = qr_translation_rule
 ad.primitive_jvps[qr_p] = qr_jvp_rule
+batching.primitive_batchers[qr_p] = qr_batching_rule
 
 
 # Singular value decomposition
@@ -411,6 +514,33 @@ def svd_abstract_eval(operand, full_matrices, compute_uv):
     vt = operand
   return core.AbstractTuple((s, u, vt))
 
+def svd_jvp_rule(primals, tangents, full_matrices, compute_uv):
+  if full_matrices:
+    #TODO: implement full matrices case, documented here: https://people.maths.ox.ac.uk/gilesm/files/NA-08-01.pdf
+    raise NotImplementedError("Singular value decomposition JVP not implemented for full matrices")
+
+  A, = primals
+  dA, = tangents
+  s, U, Vt = svd_p.bind(A, full_matrices=False, compute_uv=True)
+
+  k = s.shape[-1]
+  Ut, V = np.conj(U).T, np.conj(Vt).T
+  s_dim = s[..., None, :]
+  dS = np.dot(np.dot(Ut, dA), V)
+  ds = np.real(np.diag(dS))
+  F = 1 / (np.square(s_dim) - np.square(s_dim.T) + np.eye(k)) - np.eye(k)
+  dSS = s_dim * dS
+  SdS = s_dim.T * dS
+  dU = np.dot(U, F * (dSS + dSS.T))
+  dV = np.dot(V, F * (SdS + SdS.T))
+
+  m, n = A.shape[-2], A.shape[-1]
+  if m > n:
+    dU = dU + np.dot(np.eye(m) - np.dot(U, Ut), np.dot(dA, V)) / s_dim
+  if n > m:
+    dV = dV + np.dot(np.eye(n) - np.dot(V, Vt), np.dot(np.conj(dA).T, U)) / s_dim
+  return core.pack((s, U, Vt)), core.pack((ds, dU, dV.T))
+
 def svd_cpu_translation_rule(c, operand, full_matrices, compute_uv):
   shape = c.GetShape(operand)
   dtype = shape.element_type().type
@@ -423,8 +553,16 @@ def svd_cpu_translation_rule(c, operand, full_matrices, compute_uv):
     raise NotImplementedError(
         "Only unbatched singular value decomposition is implemented on CPU")
 
+def svd_batching_rule(batched_args, batch_dims, full_matrices, compute_uv):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.bdim_at_front(x, bd)
+  return svd_p.bind(x, full_matrices=full_matrices, compute_uv=compute_uv), 0
+
 svd_p = Primitive('svd')
 svd_p.def_impl(svd_impl)
 svd_p.def_abstract_eval(svd_abstract_eval)
 xla.translations[svd_p] = svd_translation_rule
-xla.backend_specific_translations['Host'][svd_p] = svd_cpu_translation_rule
+xla.backend_specific_translations['cpu'][svd_p] = svd_cpu_translation_rule
+ad.primitive_jvps[svd_p] = svd_jvp_rule
+batching.primitive_batchers[svd_p] = svd_batching_rule
