@@ -29,24 +29,45 @@ from distutils.util import strtobool
 
 from ..config import flags
 import numpy as onp  # 'onp' rather than 'np' to distinguish from autograd.numpy
+import six
+
+import jaxlib
+
+_minimum_jaxlib_version = (0, 1, 14)
+try:
+  from jaxlib import version as jaxlib_version
+except:
+  # jaxlib is too old to have version number.
+  msg = 'This version of jax requires jaxlib version >= {}.'
+  raise ImportError(msg.format('.'.join(map(str, _minimum_jaxlib_version))))
+
+
+# Check the jaxlib version before importing anything else from jaxlib.
+def _check_jaxlib_version():
+  version = tuple(int(x) for x in jaxlib_version.__version__.split('.'))
+  if version < _minimum_jaxlib_version:
+    msg = 'jaxlib is version {}, but this version of jax requires version {}.'
+    raise ValueError(msg.format('.'.join(map(str, version)),
+                                '.'.join(map(str, _minimum_jaxlib_version))))
+
+_check_jaxlib_version()
+
 
-from jaxlib import xla_data_pb2
 from jaxlib import xla_client
+from jaxlib import xla_data_pb2
+
+# TODO(phawkins): This is a workaround for older jaxlib versions. Remove after a
+# jaxlib release.
+try:
+  from jaxlib import xrt
+except ImportError:
+  xrt = None
 
 
 FLAGS = flags.FLAGS
 flags.DEFINE_bool('jax_enable_x64',
-                  strtobool(os.getenv('JAX_ENABLE_X64', "False")),
+                  strtobool(os.getenv('JAX_ENABLE_X64', 'False')),
                   'Enable 64-bit types to be used.')
-flags.DEFINE_string('jax_dump_hlo_graph', None, 'Regexp of HLO graphs to dump.')
-flags.DEFINE_bool('jax_hlo_profile', False, 'Enables HLO profiling mode.')
-flags.DEFINE_string('jax_dump_hlo_unoptimized', None,
-                    'Dirpath for unoptimized HLO dump.')
-flags.DEFINE_string('jax_dump_hlo_optimized', None,
-                    'Dirpath for optimized HLO dump.')
-flags.DEFINE_string('jax_dump_hlo_per_pass', None,
-                    'Dirpath for per-pass HLO dump.')
-flags.DEFINE_integer('jax_replica_count', 1, 'Replica count for computations.')
 flags.DEFINE_string(
     'jax_xla_backend', 'xla',
     'Either "xla" for the XLA service directly, or "xrt" for an XRT backend.')
@@ -54,50 +75,19 @@ flags.DEFINE_string(
     'jax_backend_target', 'local',
     'Either "local" or "rpc:address" to connect to a remote service target.')
 flags.DEFINE_string(
-    'jax_platform_name', '',
-    'Platform name for XLA. The default is to attempt to use a '
-    'GPU if available, but fall back to CPU otherwise. To set '
-    'the platform manually, pass "Host" for CPU or "CUDA" for '
-    'GPU.')
+    'jax_platform_name',
+    os.getenv('JAX_PLATFORM_NAME', ''),
+    'Platform name for XLA. The default is to attempt to use a GPU if '
+    'available, but fall back to CPU otherwise. To set the platform manually, '
+    'pass "cpu" for CPU or "gpu" for GPU.')
 
 
-# Prefix for HLO-dump flags indicating output should go into Sponge's
-# visible output files.
-_SPONGE_PREFIX = '/SPONGE/'
-
-_platform_name = None  # set to the active platform name
-
-
-def _hlo_path(path, name):
-  path = path.replace(_SPONGE_PREFIX,
-                      os.getenv('TEST_UNDECLARED_OUTPUTS_DIR', ''))
-  path = os.path.join(path, name)
-  if not os.path.exists(path):
-    os.mkdir(path)
-  return path
-
-
-def get_compile_options():
+def get_compile_options(num_replicas=None):
   """Returns the compile options to use, as derived from flag values."""
   compile_options = None
-  if FLAGS.jax_dump_hlo_graph is not None:
-    compile_options = get_xla_client().CompileOptions()
-    compile_options.generate_hlo_graph = FLAGS.jax_dump_hlo_graph
-  if FLAGS.jax_hlo_profile:
-    compile_options = compile_options or get_xla_client().CompileOptions()
-    compile_options.hlo_profile = True
-  if FLAGS.jax_dump_hlo_unoptimized:
-    compile_options = compile_options or get_xla_client().CompileOptions()
-    path = _hlo_path(FLAGS.jax_dump_hlo_unoptimized, 'hlo_unoptimized')
-    compile_options.dump_unoptimized_hlo_proto_to = path
-  if FLAGS.jax_dump_hlo_optimized:
-    compile_options = compile_options or get_xla_client().CompileOptions()
-    path = _hlo_path(FLAGS.jax_dump_hlo_optimized, 'hlo_optimized')
-    compile_options.dump_optimized_hlo_proto_to = path
-  if FLAGS.jax_dump_hlo_per_pass:
-    compile_options = compile_options or get_xla_client().CompileOptions()
-    path = _hlo_path(FLAGS.jax_dump_hlo_per_pass, 'hlo_per_pass')
-    compile_options.dump_per_pass_hlo_proto_to = path
+  if num_replicas is not None:
+    compile_options = compile_options or xla_client.CompileOptions()
+    compile_options.num_replicas = num_replicas
   return compile_options
 
 
@@ -114,85 +104,50 @@ def memoize_thunk(func):
   return lambda: cached[0] if cached else (cached.append(func()) or cached[0])
 
 
-@memoize_thunk
-def get_xla_client():
-  return _get_xla_client(FLAGS.jax_xla_backend,
-                         FLAGS.jax_platform_name,
-                         FLAGS.jax_replica_count)
-
-
-def _jaxlib_version():
-  if hasattr(xla_client, 'version'):
-    return xla_client.version()
-  else:
-    return (0, 1, 6)  # The version before xla_client.version() was added.
-
-
-def _check_jaxlib_version():
-  minimum_version = (0, 1, 6)
-  version = _jaxlib_version()
-  if version < minimum_version:
-    msg = 'jaxlib is version {}, but this version of jax requires version {}.'
-    raise ValueError(msg.format('.'.join(map(str, version)),
-                                '.'.join(map(str, minimum_version))))
-
-
-def _get_xla_client(backend_name, platform_name, replica_count):
-  """Configures and returns a handle to the XLA client.
-
-  Args:
-    backend_name: backend name, 'xla' or 'xrt'
-    platform_name: platform name for XLA backend
-    replica_count: number of computation replicas with which to configure the
-      backend library.
-
-  Returns:
-    A client library module, or an object that behaves identically to one.
-  """
-  _check_jaxlib_version()
-  global _platform_name
-  xla_client.initialize_replica_count(replica_count)
-  if backend_name == 'xla':
-    if platform_name:
-      xla_client.initialize_platform_name(platform_name)
-      _platform_name = platform_name
-    else:
-      try:
-        xla_client.initialize_platform_name('CUDA')
-        _platform_name = 'CUDA'
-      except RuntimeError:
-        warnings.warn('No GPU found, falling back to CPU.')
-        xla_client.initialize_platform_name('Host')
-        _platform_name = 'Host'
-  return xla_client
-
-
-def get_replica_count():
-  return get_xla_client().get_replica_count()
-
-
 _backends = {}
 
 def register_backend(name, factory):
   _backends[name] = factory
 
+def _get_local_backend():
+  platform = FLAGS.jax_platform_name
 
-if hasattr(xla_client, 'XlaLocalBackend'):
-  register_backend('xla', lambda: xla_client.XlaLocalBackend())
-  register_backend('xrt',
-                   lambda: xla_client.XrtBackend(FLAGS.jax_backend_target))
-else:
-  # TODO(phawkins): this case is for cross-version compatibility. Delete this
-  # case after a Jaxlib update.
-  register_backend(
-    'xla', lambda: xla_client.BackendSpec(xla_client.BackendType.XLA_LOCAL, ''))
-  register_backend(
-    'xrt', lambda: xla_client.BackendSpec(xla_client.BackendType.XRT,
-                                          FLAGS.jax_backend_target))
+  # Canonicalize platform names.
+  cpu = 'cpu'
+  gpu = 'gpu'
+  if platform == 'Host':
+    platform = cpu
+  elif platform == 'CUDA':
+    platform = gpu
+  elif platform == '':
+    platform = None
+
+  backend = xla_client.get_local_backend(platform)
+  if backend is None:
+    raise RuntimeError("No local XLA backends found.")
+
+  if backend.platform == cpu and platform != cpu:
+    warnings.warn('No GPU/TPU found, falling back to CPU.')
+
+  return backend
+
+def _get_xrt_backend():
+  # TODO(phawkins): support non-TPU devices.
+  tf_device_name = "TPU"
+  worker = "tpu_worker"
+  tf_context = xrt.get_tf_context(FLAGS.jax_backend_target, worker)
+  backend = xrt.XrtBackend(tf_context, tf_device_name)
+  #  TODO(phawkins) fix XrtBackend to set the following and remove this line.
+  backend.platform = "TPU"
+  return backend
+
+
+register_backend('xla', _get_local_backend)
+register_backend('xrt', _get_xrt_backend)
 
 
 @memoize_thunk
-def _get_backend():
+def get_backend():
   backend = _backends.get(FLAGS.jax_xla_backend)
   if backend is None:
     msg = 'Unknown jax_xla_backend value "{}".'
@@ -200,9 +155,20 @@ def _get_backend():
   return backend()
 
 
-def device_put(pyval, replica=0):
-  client = get_xla_client()
-  return client.LocalBuffer.from_pyval(pyval, replica, backend=_get_backend())
+def device_count():
+  return get_backend().device_count()
+
+
+def device_put(pyval, device_num=0):
+  return xla_client.LocalBuffer.from_pyval(pyval, device_num,
+                                           backend=get_backend())
+
+def device_put_many(pyvals_and_devices):
+  return [device_put(pyval, device) for (pyval, device) in pyvals_and_devices]
+
+def make_tuple(bufs, device_num=0):
+  return xla_client.Buffer.make_tuple(bufs, device=device_num,
+                                      backend=get_backend())
 
 
 Shape = xla_client.Shape        # pylint: disable=invalid-name
@@ -210,41 +176,15 @@ Shape = xla_client.Shape        # pylint: disable=invalid-name
 
 ### utility functions
 
-# Similar or identical dtype-to-etype conversion tables exist in the XLA
-# clients, but because their organization hasn't been made consistent across
-# clients yet, we repeat the information here.
-_etype_to_dtype = {
-    xla_data_pb2.PRED: onp.dtype('bool'),
-    xla_data_pb2.S8: onp.dtype('int8'),
-    xla_data_pb2.S16: onp.dtype('int16'),
-    xla_data_pb2.S32: onp.dtype('int32'),
-    xla_data_pb2.S64: onp.dtype('int64'),
-    xla_data_pb2.U8: onp.dtype('uint8'),
-    xla_data_pb2.U16: onp.dtype('uint16'),
-    xla_data_pb2.U32: onp.dtype('uint32'),
-    xla_data_pb2.U64: onp.dtype('uint64'),
-    xla_data_pb2.F16: onp.dtype('float16'),
-    xla_data_pb2.F32: onp.dtype('float32'),
-    xla_data_pb2.F64: onp.dtype('float64'),
-    xla_data_pb2.C64: onp.dtype('complex64'),
-    xla_data_pb2.C128: onp.dtype('complex128'),
-}
-
-# Note the conversion on the key. Numpy has a known issue wherein dtype hashing
-# doesn't work as expected (https://github.com/numpy/numpy/issues/7242). Thus,
-# when keying by dtype in this dict, we use the string form of dtypes.
-_dtype_to_etype = {str(dt): et for et, dt in _etype_to_dtype.items()}
-
-
 @memoize
 def dtype_to_etype(dtype):
   """Convert from dtype to canonical etype (reading FLAGS.jax_enable_x64)."""
-  return _dtype_to_etype[canonicalize_dtype(dtype)]
+  return xla_client.DTYPE_TO_XLA_ELEMENT_TYPE[canonicalize_dtype(dtype)]
 
 @memoize
 def dtype_to_etype_exact(dtype):
   """Convert from dtype to exact etype (ignoring FLAGS.jax_enable_x64)."""
-  return _dtype_to_etype[str(onp.dtype(dtype))]
+  return xla_client.dtype_to_etype(dtype)
 
 
 _dtype_to_32bit_dtype = {
@@ -268,7 +208,8 @@ def canonicalize_dtype(dtype):
 
 @memoize_thunk
 def supported_numpy_dtypes():
-  return {canonicalize_dtype(dtype) for dtype in _etype_to_dtype.values()}
+  return {canonicalize_dtype(dtype)
+          for dtype in xla_client.XLA_ELEMENT_TYPE_TO_DTYPE.values()}
 
 
 def canonicalize_shape(shape):
@@ -307,48 +248,36 @@ def shape_of(value):
 
 def infeed_put(replica_id, pyval):
   pyval = normalize_to_xla_dtypes(pyval)
-  return get_xla_client().transfer_to_infeed(
+  return xla_client.transfer_to_infeed(
       pyval, replica_number=replica_id)
 
 
-class _JaxComputationBuilderBase(object):
+class _JaxComputationBuilder(xla_client.ComputationBuilder):
   """Base class implementing all of JaxComputationBuilder.
 
   This class is intended to override and augment the interface of an XLA
-  ComputationBuilder to form JaxComputationBuilder, as made clear by
-  `get_jax_computation_builder_class`, which relies on Python's
-  method-resolution order to set up inheritance-like behavior. The class
-  inheritance setup is deferred because the choice of the XLA ComputationBuilder
-  class is based on the result of `get_xla_client()`. That is, the choice is
-  based at least on the setting of flags, which are available only after module
-  initialization time.
+  ComputationBuilder to form JaxComputationBuilder
   """
-  # The JAXComputationBuilder is implemented using subclassing and inheritance
-  # (via this base class), rather than a wrap-and-delegate style, simply to
-  # avoid having to spell out all the methods to be forwarded to a wrapped
-  # ComputationBuilder, especially since the underlying ComputationBuilders are
-  # likely to be revised in the future. An alternative is to generate these
-  # forwarding methods programmatically.
 
   # Method name case follows that of the XLA ComputationBuilder
   # pylint: disable=invalid-name
 
   def Build(self, *args, **kwargs):
-    return super(_JaxComputationBuilderBase, self).Build(
-        *args, backend=_get_backend(), **kwargs)
+    return super(_JaxComputationBuilder, self).Build(
+        *args, **kwargs)
 
   def Parameter(self, value, name=None, parameter_num=None):
-    return super(_JaxComputationBuilderBase, self).ParameterWithShape(
+    return super(_JaxComputationBuilder, self).ParameterWithShape(
         shape_of(value), name=name, parameter_num=parameter_num)
 
   def NumpyArrayConstant(self, value, canonicalize_types=True):
     if canonicalize_types:
       value = normalize_to_xla_dtypes(value)
-    return super(_JaxComputationBuilderBase, self).Constant(value)
+    return super(_JaxComputationBuilder, self).Constant(value)
 
   def ConstantLike(self, example_value, value, canonicalize_types=True):
     example_value = onp.asarray(example_value)
-    return self.Constant(onp.array(value).astype(example_value.dtype))
+    return self.Constant(onp.array(value, dtype=example_value.dtype))
 
   def Constant(self, py_val, canonicalize_types=True):
     """Translate constant `py_val` to a constant for this ComputationBuilder.
@@ -365,16 +294,18 @@ class _JaxComputationBuilderBase(object):
     else:
       raise TypeError("No constant handler for type: {}".format(py_type))
 
-
-@memoize_thunk
-def get_jax_computation_builder_class():
-  xla_base = get_xla_client().ComputationBuilder
-  jax_base = _JaxComputationBuilderBase
-  return type('JaxComputationBuilder', (jax_base, xla_base), {})
+  # TODO(mattjj): remove when CRS is added to XLA:CPU
+  def CrossReplicaSum(self, operand, replica_groups):
+    """Workaround for CrossReplicaSum not being implemented on some backends."""
+    if len(replica_groups[0]) == 1:
+      return operand
+    else:
+      return super(_JaxComputationBuilder, self).CrossReplicaSum(operand,
+                                                                 replica_groups)
 
 
 def make_computation_builder(name):
-  return get_jax_computation_builder_class()(name)
+  return _JaxComputationBuilder(name)
 
 
 def register_constant_handler(type_, handler_fun):
@@ -422,5 +353,8 @@ def _scalar_constant_handler(c, val, canonicalize_types=True):
 for scalar_type in [onp.int8, onp.int16, onp.int32, onp.int64,
                     onp.uint8, onp.uint16, onp.uint32, onp.uint64,
                     onp.float16, onp.float32, onp.float64, onp.float128,
-                    float, int, bool, onp.bool_]:
+                    float, int, bool, onp.bool_, onp.longlong]:
   register_constant_handler(scalar_type, _scalar_constant_handler)
+
+if six.PY2:
+  register_constant_handler(long, _scalar_constant_handler)
