@@ -21,6 +21,7 @@ import itertools
 import re
 import string
 import warnings
+import types
 
 import numpy as onp
 import opt_einsum
@@ -34,6 +35,7 @@ from ..interpreters.xla import DeviceArray
 from .. import lax
 from ..util import memoize, partial, get_module_functions, unzip2, prod as _prod
 from ..lib import xla_bridge
+from ..lib.xla_bridge import xla_client
 
 if six.PY3:
   def removechars(s, chars):
@@ -90,7 +92,7 @@ result_type = onp.result_type
 shape = _shape = onp.shape
 ndim = _ndim = onp.ndim
 size = onp.size
-_dtype = lax._dtype
+_dtype = lax.dtype
 
 bool_ = onp.bool_
 uint8 = onp.uint8
@@ -115,6 +117,8 @@ inexact = onp.inexact
 complexfloating = onp.complexfloating
 floating = onp.floating
 integer = onp.integer
+signedinteger = onp.signedinteger
+unsignedinteger = onp.unsignedinteger
 
 iinfo = onp.iinfo
 finfo = onp.finfo
@@ -146,17 +150,11 @@ def _promote_dtypes(*args):
   if len(args) < 2:
     return args
   else:
-    from_dtypes = (x if type(x) in _builtin_numeric_types else _dtype(x)
-                   for x in args)
+    from_dtypes = map(_dtype, args)
     to_dtype = xla_bridge.canonicalize_dtype(result_type(*from_dtypes))
     return [lax.convert_element_type(x, to_dtype)
             if _dtype(x) != to_dtype else x for x in args]
 
-if six.PY3:
-  _builtin_numeric_types = (int, float, complex)
-else:
-  _builtin_numeric_types = (int, float, long, complex)
-
 def _promote_to_result_dtype(op, *args):
   """Convenience function to promote args directly to the op's result dtype."""
   to_dtype = _result_dtype(op, *args)
@@ -284,10 +282,10 @@ equal = _one_to_one_binop(onp.equal, lax.eq)
 multiply = _one_to_one_binop(onp.multiply, lax.mul)
 not_equal = _one_to_one_binop(onp.not_equal, lax.ne)
 subtract = _one_to_one_binop(onp.subtract, lax.sub)
-power = _one_to_one_binop(onp.power, lax.pow, True)
 arctan2 = _one_to_one_binop(onp.arctan2, lax.atan2, True)
 minimum = _one_to_one_binop(onp.minimum, lax.min)
 maximum = _one_to_one_binop(onp.maximum, lax.max)
+float_power = _one_to_one_binop(onp.float_power, lax.pow, True)
 
 
 def _comparison_op(numpy_fn, lax_fn):
@@ -388,6 +386,28 @@ def _float_divmod(x1, x2):
   return lax.round(div), mod
 
 
+@_wraps(onp.power)
+def power(x1, x2):
+  x1 = asarray(x1)
+  x2 = asarray(x2)
+  x1, x2 = _promote_args_like(onp.power, x1, x2)
+  dtype = _dtype(x1)
+  if not issubdtype(dtype, integer):
+    return lax.pow(x1, x2)
+
+  # Integer power => use binary exponentiation.
+
+  # TODO(phawkins): add integer pow support to XLA.
+  bits = 6  # Anything more would overflow for any x1 > 1
+  acc = ones(shape(x1), dtype=dtype)
+  for _ in xrange(bits):
+    acc = where(lax.bitwise_and(x2, _constant_like(x2, 1)),
+                lax.mul(acc, x1), acc)
+    x1 = lax.mul(x1, x1)
+    x2 = lax.shift_right_logical(x2, _constant_like(x2, 1))
+  return acc
+
+
 @_wraps(onp.logaddexp)
 def logaddexp(x1, x2):
   x1, x2 = _promote_shapes(*_promote_to_result_dtype(onp.logaddexp, x1, x2))
@@ -430,6 +450,12 @@ mod = remainder
 fmod = _wraps(onp.fmod)(lambda x, y: lax.rem(x, y))
 
 
+@_wraps(onp.cbrt)
+def cbrt(x):
+  x, = _promote_to_result_dtype(onp.cbrt, x)
+  return lax.sign(x) * power(lax.abs(x), _constant_like(x, 1. / 3.))
+
+
 @_wraps(onp.sqrt)
 def sqrt(x):
   x, = _promote_to_result_dtype(onp.sqrt, x)
@@ -571,21 +597,58 @@ def angle(x):
   return lax.atan2(im, re)
 
 
+@_wraps(onp.diff)
+def diff(a, n=1, axis=-1,):
+  if not isinstance(a, ndarray) or a.ndim == 0:
+    return a
+  if n == 0:
+    return a
+  if n < 0:
+    raise ValueError(
+      "order must be non-negative but got " + repr(n))
+
+  nd = a.ndim
+
+  slice1 = [slice(None)] * nd
+  slice2 = [slice(None)] * nd
+  slice1[axis] = slice(1, None)
+  slice2[axis] = slice(None, -1)
+  slice1 = tuple(slice1)
+  slice2 = tuple(slice2)
+
+  op = not_equal if a.dtype == onp.bool_ else subtract
+  for _ in range(n):
+    a = op(a[slice1], a[slice2])
+
+  return a
+
+
+@_wraps(onp.isrealobj)
+def isrealobj(a):
+  return not iscomplexobj(a)
+
+
 @_wraps(onp.reshape)
-def reshape(a, newshape, order="C"):  # pylint: disable=missing-docstring
-  if order == "C" or order is None:
-    dims = None
+def reshape(a, newshape, order="C"):
+  try:
+    return a.reshape(newshape, order=order)
+  except AttributeError:
+    return _reshape(a, newshape, order=order)
+
+def _reshape(a, newshape, order="C"):
+  dummy_val = onp.broadcast_to(0, shape(a))  # zero strides
+  computed_newshape = onp.reshape(dummy_val, newshape).shape
+
+  if order == "C":
+    return lax.reshape(a, computed_newshape, None)
   elif order == "F":
     dims = onp.arange(ndim(a))[::-1]
+    return lax.reshape(a, computed_newshape[::-1], dims).T
   elif order == "A":
     raise NotImplementedError("np.reshape order=A is not implemented.")
   else:
     raise ValueError("Unexpected value for 'order' argument: {}.".format(order))
 
-  dummy_val = onp.broadcast_to(0, shape(a))  # zero strides
-  computed_newshape = onp.reshape(dummy_val, newshape).shape
-  return lax.reshape(a, computed_newshape, dims)
-
 
 @_wraps(onp.ravel)
 def ravel(a, order="C"):
@@ -643,12 +706,25 @@ def isclose(a, b, rtol=1e-05, atol=1e-08):
       dtype = _result_dtype(real, a)
     rtol = lax.convert_element_type(rtol, dtype)
     atol = lax.convert_element_type(atol, dtype)
-    return lax.le(
+    out = lax.le(
       lax.abs(lax.sub(a, b)),
       lax.add(atol, lax.mul(rtol, lax.abs(b))))
+    return _maybe_numpy_1_13_isclose_behavior(a, out)
   else:
     return lax.eq(a, b)
 
+numpy_version = tuple(map(int, onp.version.version.split('.')))
+if numpy_version < (1, 14):
+  # see discussion at https://github.com/numpy/numpy/pull/9720
+  def _maybe_numpy_1_13_isclose_behavior(a, out):
+    if size(out) == 1 and issubdtype(_dtype(a), complexfloating):
+      return lax.reshape(out, (1,))
+    else:
+      return out
+else:
+  def _maybe_numpy_1_13_isclose_behavior(a, out):
+    return out
+
 
 @_wraps(onp.where)
 def where(condition, x=None, y=None):
@@ -761,10 +837,12 @@ def round(a, decimals=0):
 around = round
 
 
-# Caution: If fast math mode is enabled, the semantics of inf and nan are not
-# preserved by XLA/LLVM, and the behavior of inf/nan values is unpredictable.
-# To disable fast math mode on CPU, set the environment variable
-# XLA_FLAGS=--xla_cpu_enable_fast_math=false.
+@_wraps(onp.fix)
+def fix(x, out=None):
+  if out is not None:
+    raise ValueError("fix does not support the `out` argument.")
+  zero = lax._const(x, 0)
+  return where(lax.ge(x, zero), lax.floor(x), lax.ceil(x))
 
 @_wraps(onp.isfinite)
 def isfinite(x):
@@ -780,17 +858,19 @@ def isfinite(x):
 def isinf(x):
   dtype = _dtype(x)
   if issubdtype(dtype, floating):
-    return lax.eq(lax.abs(x), inf)
+    return lax.eq(lax.abs(x), _constant_like(x, inf))
   elif issubdtype(dtype, complexfloating):
-    return lax.bitwise_or(lax.eq(lax.abs(real(x)), inf),
-                          lax.eq(lax.abs(imag(x)), inf))
+    re = lax.real(x)
+    im = lax.imag(x)
+    return lax.bitwise_or(lax.eq(lax.abs(re), _constant_like(re, inf)),
+                          lax.eq(lax.abs(im), _constant_like(im, inf)))
   else:
     return full_like(x, False, dtype=bool_)
 
 def _isposneginf(infinity, x):
   dtype = _dtype(x)
   if issubdtype(dtype, floating):
-    return lax.eq(x, infinity)
+    return lax.eq(x, _constant_like(x, infinity))
   elif issubdtype(dtype, complexfloating):
     raise ValueError("isposinf/isneginf are not well defined for complex types")
   else:
@@ -807,9 +887,10 @@ def isnan(x):
 @_wraps(onp.nan_to_num)
 def nan_to_num(x, copy=True):
   del copy
-  if iscomplexobj(x):
-    raise ValueError("nan_to_num is not well defined for complex types")
-  info = finfo(xla_bridge.canonicalize_dtype(_dtype(x)))
+  dtype = _dtype(x)
+  if issubdtype(dtype, complexfloating):
+    return lax.complex(nan_to_num(lax.real(x)), nan_to_num(lax.imag(x)))
+  info = finfo(xla_bridge.canonicalize_dtype(dtype))
   x = where(isnan(x), _constant_like(x, 0), x)
   x = where(isposinf(x), _constant_like(x, info.max), x)
   x = where(isneginf(x), _constant_like(x, info.min), x)
@@ -887,10 +968,58 @@ def mean(a, axis=None, dtype=None, out=None, keepdims=False):
     else:
       dtype = _dtype(a)
 
-  td = true_divide(
+  return lax.div(
       sum(a, axis, dtype=dtype, keepdims=keepdims),
       lax.convert_element_type(normalizer, dtype))
-  return lax.convert_element_type(td, dtype)
+
+@_wraps(onp.average)
+def average(a, axis=None, weights=None, returned=False):
+    a = asarray(a)
+
+    if weights is None: # Treat all weights as 1
+        avg = mean(a, axis=axis)
+        if axis is None:
+            weights_sum = full((), size(a), dtype=avg.dtype)
+        else:
+            weights_sum = full_like(avg, a.shape[axis], dtype=avg.dtype)
+    else:
+        weights = asarray(weights)
+
+        if issubdtype(a.dtype, integer) or issubdtype(a.dtype, bool_):
+            out_dtype = xla_bridge.canonicalize_dtype(result_type(a.dtype,
+                                                                  weights.dtype,
+                                                                  floating))
+        else:
+            out_dtype = xla_bridge.canonicalize_dtype(result_type(a.dtype, weights.dtype))
+
+        a_shape = shape(a)
+        a_ndim = len(a_shape)
+        weights_shape = shape(weights)
+        axis = None if axis is None else _canonicalize_axis(axis, a_ndim)
+
+        if a_shape != weights_shape:
+            # Make sure the dimensions work out
+            if axis is None:
+                raise ValueError("Axis must be specified when shapes of a and "
+                                 "weights differ.")
+            if len(weights_shape) != 1:
+                raise ValueError("1D weights expected when shapes of a and "
+                                 "weights differ.")
+            if weights_shape[0] != a_shape[axis]:
+                raise ValueError("Length of weights not "
+                                 "compatible with specified axis.")
+
+            weights = broadcast_to(weights, (a_ndim - 1) * (1,) + weights_shape)
+            weights = moveaxis(weights, -1, axis)
+
+        weights_sum = sum(weights, axis=axis, dtype=out_dtype)
+        avg = sum(multiply(a, weights), axis=axis, dtype=out_dtype) / weights_sum
+
+    if returned:
+        if avg.shape != weights_sum.shape:
+            weights_sum = broadcast_to(weights_sum, avg.shape)
+        return avg, weights_sum
+    return avg
 
 
 @_wraps(onp.var)
@@ -990,7 +1119,7 @@ def _make_cumulative_reduction(onp_reduction, window_reduce, init_val,
     window_dims = [1] * num_dims
     window_dims[axis] = a_shape[axis]
     return window_reduce(
-       a, window_dims, strides, xla_bridge.get_xla_client().PaddingType.VALID)
+       a, window_dims, strides, xla_client.PaddingType.VALID)
 
   return cumulative_reduction
 
@@ -1028,10 +1157,11 @@ def pad(array, pad_width, mode, constant_values=0):
     raise NotImplementedError(msg.format(mode))
 
   array = asarray(array)
-  pad_width = onp.broadcast_to(onp.asarray(pad_width), (array.ndim, 2))
-  constant_values = broadcast_to(asarray(constant_values), (array.ndim, 2))
-  for i in xrange(array.ndim):
-    widths = [(0, 0, 0)] * array.ndim
+  nd = ndim(array)
+  pad_width = onp.broadcast_to(onp.asarray(pad_width), (nd, 2))
+  constant_values = broadcast_to(asarray(constant_values), (nd, 2))
+  for i in xrange(nd):
+    widths = [(0, 0, 0)] * nd
     widths[i] = (pad_width[i, 0], 0, 0)
     array = lax.pad(array, constant_values[i, 0], widths)
     widths[i] = (0, pad_width[i, 1], 0)
@@ -1054,6 +1184,15 @@ def stack(arrays, axis=0):
     new_arrays.append(reshape(a, new_shape))
   return concatenate(new_arrays, axis=axis)
 
+@_wraps(onp.tile)
+def tile(a, reps):
+  if isinstance(reps, int):
+    reps = (reps,)
+  a = reshape(a, (1,) * (len(reps) - ndim(a)) + shape(a))
+  reps = (1,) * (ndim(a) - len(reps)) + reps
+  for i, rep in enumerate(reps):
+    a = concatenate([a] * rep, axis=i)
+  return a
 
 @_wraps(onp.concatenate)
 def concatenate(arrays, axis=0):
@@ -1098,7 +1237,7 @@ def column_stack(tup):
 def atleast_1d(*arys):
   if len(arys) == 1:
     arr = array(arys[0])
-    return arr if arr.ndim >= 1 else arr.reshape(-1)
+    return arr if ndim(arr) >= 1 else reshape(arr, -1)
   else:
     return [atleast_1d(arr) for arr in arys]
 
@@ -1107,7 +1246,7 @@ def atleast_1d(*arys):
 def atleast_2d(*arys):
   if len(arys) == 1:
     arr = array(arys[0])
-    return arr if arr.ndim >= 2 else arr.reshape((1, -1))
+    return arr if ndim(arr) >= 2 else reshape(arr, (1, -1))
   else:
     return [atleast_2d(arr) for arr in arys]
 
@@ -1117,32 +1256,31 @@ def atleast_3d(*arys):
   if len(arys) == 1:
     arr = array(arys[0])
     if ndim(arr) <= 1:
-      arr = arr.reshape((1, -1, 1))
+      arr = reshape(arr, (1, -1, 1))
     elif ndim(arr) == 2:
-      arr = arr.reshape(shape(arr) + (1,))
+      arr = reshape(arr, shape(arr) + (1,))
     return arr
   else:
     return [atleast_3d(arr) for arr in arys]
 
 
-# TODO(mattjj): can this be simplified?
 @_wraps(onp.array)
 def array(object, dtype=None, copy=True, order="K", ndmin=0):
   del copy  # Unused.
   if ndmin != 0 or order != "K":
     raise NotImplementedError("Only implemented for order='K', ndmin=0.")
 
-  if hasattr(object, '__asarray__'):
-    return object.__asarray__(dtype)
-  elif isinstance(object, ndarray):
+  if isinstance(object, ndarray):
     if dtype and _dtype(object) != dtype:
       return lax.convert_element_type(object, dtype)
     else:
       return object
+  elif hasattr(object, '__array__'):
+    # this case is for duck-typed handling of objects that implement `__array__`
+    return array(object.__array__(), dtype)
   elif isinstance(object, (list, tuple)):
     if object:
-      subarrays = [expand_dims(array(elt, dtype=dtype), 0) for elt in object]
-      return concatenate(subarrays)
+      return stack([array(elt, dtype=dtype) for elt in object])
     else:
       return onp.array([], dtype)
   elif isscalar(object):
@@ -1178,6 +1316,8 @@ def full_like(a, fill_value, dtype=None):
 
 @_wraps(onp.zeros)
 def zeros(shape, dtype=onp.dtype("float64")):
+  if isinstance(shape, types.GeneratorType):
+    raise TypeError("expected sequence object with len >= 0 or a single integer")
   shape = (shape,) if onp.isscalar(shape) else shape
   return lax.full(shape, 0, dtype)
 
@@ -1187,6 +1327,15 @@ def ones(shape, dtype=onp.dtype("float64")):
   return lax.full(shape, 1, dtype)
 
 
+@_wraps(onp.array_equal)
+def array_equal(a1, a2):
+  try:
+    a1, a2 = asarray(a1), asarray(a2)
+  except Exception:
+    return False
+  return shape(a1) == shape(a2) and all(asarray(a1 == a2))
+
+
 # We can't create uninitialized arrays in XLA; use zeros for empty.
 empty_like = zeros_like
 empty = zeros
@@ -1217,16 +1366,18 @@ def identity(n, dtype=None):
 
 @_wraps(onp.arange)
 def arange(*args, **kwargs):
-  # attempt to generate a lazy IotaConstant, otherwise fall back to raw numpy
-  # TODO(mattjj): add tests for this function, then re-enable
-  # dtype = kwargs.pop("dtype", None)
-  # if not args:
-  #   raise TypeError("Required argument 'start' (pos 1) not found")  # same as numpy error
-  # elif len(args) == 1 and not kwargs:
-  #   stop, = args
-  #   dtype = dtype or _dtype(stop)
-  #   if onp.issubdtype(dtype, onp.integer):
-  #     return lax.iota(dtype, stop)  # avoids materializing
+  dtype = kwargs.pop("dtype", None)
+  if not args:
+    raise TypeError("Required argument 'start' (pos 1) not found")  # same as numpy error
+
+  # If called like np.arange(N), we create a lazy lax._IotaConstant.
+  if len(args) == 1 and not kwargs:
+    stop, = args
+    dtype = dtype or _dtype(stop)
+    if onp.issubdtype(dtype, onp.integer):
+      return lax.iota(dtype, stop)  # avoids materializing
+
+  # Fall back to instantiating an ndarray in host memory
   return onp.arange(*args, **kwargs)
 
 linspace = onp.linspace
@@ -1293,6 +1444,9 @@ def trace(a, offset=0, axis1=0, axis2=1, dtype=None, out=None):
   if out:
     raise NotImplementedError("The 'out' argument to trace is not supported.")
 
+  axis1 = axis1 % ndim(a)
+  axis2 = axis2 % ndim(a)
+
   a_shape = shape(a)
   if dtype is None:
     dtype = _dtype(a)
@@ -1386,13 +1540,13 @@ def dot(a, b):  # pylint: disable=missing-docstring
     return lax.mul(a, b)
   if _max(a_ndim, b_ndim) <= 2:
     return lax.dot(a, b)
-  a_reshaped = reshape(a, (-1, shape(a)[-1]))
-  if _ndim(b) in {1, 2}:
-    out = lax.dot(a_reshaped, b)
+
+  if b_ndim == 1:
+    contract_dims = ((a_ndim - 1,), (0,))
   else:
-    b_reshaped = reshape(moveaxis(b, -2, 0), (shape(b)[-2], -1))
-    out = lax.dot(a_reshaped, b_reshaped)
-  return lax.reshape(out, a.shape[:-1] + b.shape[:-2] + b.shape[-2:][1:])
+    contract_dims = ((a_ndim - 1,), (b_ndim - 2,))
+  batch_dims = ((), ())
+  return lax.dot_general(a, b, (contract_dims, batch_dims))
 
 
 @_wraps(onp.matmul)
@@ -1434,11 +1588,16 @@ def tensordot(a, b, axes=2):
     raise TypeError(msg.format(ndim(a), ndim(b)))
 
   if type(axes) is int:
-    a, b = _promote_dtypes(a, b)
-    a_reshape = lax.reshape(a, (_prod(a.shape[:-axes]), _prod(a.shape[-axes:])))
-    b_reshape = lax.reshape(b, (_prod(b.shape[:axes]), _prod(b.shape[axes:])))
-    out_reshape = lax.dot(a_reshape, b_reshape)
-    return lax.reshape(out_reshape, a.shape[:-axes] + b.shape[axes:])
+    if axes == 0:
+      a, b = _promote_dtypes(a, b)
+      return lax.mul(lax.reshape(a, shape(a) + (1,) * ndim(b)),
+                     lax.reshape(b, (1,) * ndim(a) + shape(b)))
+    else:
+      a, b = _promote_dtypes(a, b)
+      a_reshape = lax.reshape(a, (_prod(a.shape[:-axes]), _prod(a.shape[-axes:])))
+      b_reshape = lax.reshape(b, (_prod(b.shape[:axes]), _prod(b.shape[axes:])))
+      out_reshape = lax.dot(a_reshape, b_reshape)
+      return lax.reshape(out_reshape, a.shape[:-axes] + b.shape[axes:])
   elif type(axes) in (list, tuple) and len(axes) == 2:
     ax1, ax2 = axes
     if type(ax1) == type(ax2) == int:
@@ -1459,13 +1618,23 @@ def tensordot(a, b, axes=2):
 
 
 @_wraps(onp.einsum)
-def einsum(*operands):
+def einsum(*operands, **kwargs):
+  optimize = kwargs.pop('optimize', 'auto')
+  optimize = 'greedy' if optimize is True else optimize
+  if kwargs:
+    msg = 'invalid keyword arguments for einsum: {}'
+    raise TypeError(msg.format(', '.join(kwargs)))
   # using einsum_call=True here is an internal api for opt_einsum
   operands, contractions = opt_einsum.contract_path(
-      *operands, einsum_call=True, use_blas=True)
+      *operands, einsum_call=True, use_blas=True, optimize=optimize)
   contractions = tuple(data[:3] for data in contractions)
   return _einsum(operands, contractions)
 
+@_wraps(onp.einsum_path)
+def einsum_path(subscripts, *operands, **kwargs):
+  optimize = kwargs.pop('optimize', 'greedy')
+  # using einsum_call=True here is an internal api for opt_einsum
+  return opt_einsum.contract_path(subscripts, *operands, optimize=optimize)
 
 @partial(jit, static_argnums=(1,))
 def _einsum(operands, contractions):
@@ -1534,7 +1703,7 @@ def _einsum(operands, contractions):
       batch_names = (set(lhs_names) & set(rhs_names)) - contracted_names
       lhs_batch, rhs_batch = unzip2((lhs_names.find(n), rhs_names.find(n))
                                     for n in batch_names)
-      
+
       # NOTE(mattjj): this can fail non-deterministically in python3, maybe
       # due to opt_einsum
       assert _all(name in lhs_names and name in rhs_names and
@@ -1651,40 +1820,60 @@ def outer(a, b, out=None):
     raise NotImplementedError("The 'out' argument to outer is not supported.")
   return ravel(a)[:, None] * ravel(b)
 
+@_wraps(onp.cross)
+def cross(a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
+    if axis is not None:
+        axisa = axis
+        axisb = axis
+        axisc = axis
+
+    a_ndims = len(shape(a))
+    b_ndims = len(shape(b))
+    axisa = _canonicalize_axis(axisa, a_ndims)
+    axisb = _canonicalize_axis(axisb, b_ndims)
+    a = moveaxis(a, axisa, -1)
+    b = moveaxis(b, axisb, -1)
+    a_shape = shape(a)
+    b_shape = shape(b)
+
+    if a_shape[-1] not in (2, 3) or b_shape[-1] not in (2, 3):
+        raise ValueError("Dimension must be either 2 or 3 for cross product")
+
+    if a_shape[-1] == 2 and b_shape[-1] == 2:
+        return a[..., 0] * b[..., 1] - a[..., 1] * b[..., 0]
+
+    if a_shape[-1] == 2:
+        a = concatenate((a, zeros(a_shape[:-1] + (1,), dtype=a.dtype)), axis=-1)
+    elif b_shape[-1] == 2:
+        b = concatenate((b, zeros(b_shape[:-1] + (1,), dtype=b.dtype)), axis=-1)
+
+    a0 = a[..., 0]
+    a1 = a[..., 1]
+    a2 = a[..., 2]
+    b0 = b[..., 0]
+    b1 = b[..., 1]
+    b2 = b[..., 2]
+
+    c = array([a1 * b2 - a2 * b1,
+               a2 * b0 - a0 * b2,
+               a0 * b1 - a1 * b0])
+
+    c_ndims = len(shape(c))
+    axisc = _canonicalize_axis(axisc, c_ndims)
+
+    return moveaxis(c, 0, axisc)
 
 @_wraps(onp.kron)
 def kron(a, b):
-  a_shape = shape(a)
-  b_shape = shape(b)
-  a_ndims = len(a_shape)
-  b_ndims = len(b_shape)
-  a = array(a)
-  b = array(b)
-  d = _min(a_ndims, b_ndims)
-  if d == 0:
-    return a * b
-  a_broadcast_dims = list(range(a_ndims - d, a_ndims + d, 2))
-  a_broadcast_shape = onp.ones(a_ndims + d, dtype=onp.int64)
-  a_broadcast_shape[:-2*d] = a_shape[:-d]
-  a_broadcast_shape[a_broadcast_dims] = a_shape[-d:]
-
-  b_broadcast_dims = list(range(b_ndims -d + 1, b_ndims + d + 1, 2))
-  b_broadcast_shape = onp.ones(b_ndims + d, dtype=onp.int64)
-  b_broadcast_shape[:-2*d] = b_shape[:-d]
-  b_broadcast_shape[b_broadcast_dims] = b_shape[-d:]
-
-  if a_ndims > b_ndims:
-    out_shape = onp.array(a_shape, dtype=onp.int64)
-    out_shape[-d:] *= onp.array(b_shape, dtype=onp.int64)
-  else:
-    out_shape = onp.array(b_shape, dtype=onp.int64)
-    out_shape[-d:] *= onp.array(a_shape, dtype=onp.int64)
-
-  a_broadcast = lax.broadcast_in_dim(
-    a, a_broadcast_shape, list(range(a_ndims - d)) + a_broadcast_dims)
-  b_broadcast = lax.broadcast_in_dim(
-    b, b_broadcast_shape, list(range(b_ndims - d)) + b_broadcast_dims)
-  return lax.reshape(a_broadcast * b_broadcast, out_shape)
+  a, b = _promote_dtypes(a, b)
+  if ndim(a) < ndim(b):
+    a = reshape(a, (1,) * (ndim(b) - ndim(a)) + shape(a))
+  elif ndim(b) < ndim(a):
+    b = reshape(b, (1,) * (ndim(a) - ndim(b)) + shape(b))
+  a_reshaped = reshape(a, [i for d in shape(a) for i in (d, 1)])
+  b_reshaped = reshape(b, [i for d in shape(b) for i in (1, d)])
+  out_shape = tuple(onp.multiply(shape(a), shape(b)))
+  return reshape(lax.mul(a_reshaped, b_reshaped), out_shape)
 
 
 @_wraps(onp.vander)
@@ -1763,6 +1952,41 @@ def argsort(a, axis=-1, kind='quicksort', order=None):
     return perm
 
 
+@_wraps(onp.roll)
+def roll(a, shift, axis=None):
+  a = asarray(a)
+  a_shape = shape(a)
+  if axis is None:
+    return lax.reshape(roll(ravel(a), shift, axis=0), a_shape)
+
+  a_ndim = len(a_shape)
+  if isinstance(shift, tuple):
+    if isinstance(axis, tuple):
+      if len(axis) != len(shift):
+        msg = "Mismatched lengths between shift ({}) and axis ({}) for np.roll."
+        raise ValueError(msg.format(len(shift), len(axis)))
+      axis = tuple(a for a in axis)
+    else:
+      axis = (axis,) * len(shift)
+  elif isinstance(axis, tuple):
+    shift = (shift,) * len(axis)
+  else:
+    shift = (shift,)
+    axis = (axis,)
+
+  for offset, i in zip(shift, axis):
+    i = _canonicalize_axis(i, a_ndim)
+    offset = offset % (a_shape[i] or 1)
+    slices = [slice(None)] * a_ndim
+    slices[i] = slice(None, -offset)
+    before = a[tuple(slices)]
+    slices[i] = slice(-offset, None)
+    after = a[tuple(slices)]
+    a = lax.concatenate((after, before), i)
+
+  return a
+
+
 @_wraps(onp.take)
 def take(a, indices, axis=None, out=None, mode=None):
   if out:
@@ -1792,9 +2016,8 @@ def take(a, indices, axis=None, out=None, mode=None):
       list(range(axis)) +
       list(range(axis + index_dims, len(a.shape) + index_dims - 1))),
     collapsed_slice_dims=(axis,),
-    start_index_map=(axis,),
-    index_vector_dim=index_dims)
-  return lax.gather(a, indices, dimension_numbers=dnums,
+    start_index_map=(axis,))
+  return lax.gather(a, indices[..., None], dimension_numbers=dnums,
                     slice_sizes=tuple(slice_sizes))
 
 
@@ -1955,9 +2178,7 @@ def _is_slice_none(idx):
 def _is_advanced_int_indexer(idx):
   """Returns True if idx should trigger int array indexing, False otherwise."""
   # https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#advanced-indexing
-  if isinstance(idx, (tuple, list)):
-    # We assume this check comes *after* the check for non-advanced tuple index,
-    # and hence we already know at least one element is a sequence if it's a tuple
+  if isinstance(idx, (tuple, list)) and _any(onp.ndim(elt) != 0 for elt in idx):
     return _all(e is None or e is Ellipsis or isinstance(e, slice)
                 or _is_int_arraylike(e) for e in idx)
   else:
@@ -2023,6 +2244,32 @@ hanning = onp.hanning
 kaiser = onp.kaiser  # TODO: lower via lax to allow non-constant beta.
 
 
+@_wraps(getattr(onp, "gcd", None))
+def gcd(x1, x2):
+  if (not issubdtype(_dtype(x1), integer) or
+      not issubdtype(_dtype(x2), integer)):
+    raise ValueError("Arguments to gcd must be integers.")
+  def cond_fn(xs):
+    x1, x2 = xs
+    return any(x2 != 0)
+  def body_fn(xs):
+    x1, x2 = xs
+    x1, x2 = (where(x2 != 0, x2, x1),
+              where(x2 != 0, lax.rem(x1, x2), lax._const(x2, 0)))
+    return (where(x1 < x2, x2, x1), where(x1 < x2, x1, x2))
+  x1, x2 = _promote_dtypes(lax.abs(x1), lax.abs(x2))
+  x1, x2 = broadcast_arrays(x1, x2)
+  gcd, _ = lax.while_loop(cond_fn, body_fn, (x1, x2))
+  return gcd
+
+
+@_wraps(getattr(onp, "lcm", None))
+def lcm(x1, x2):
+  d = gcd(x1, x2)
+  return where(d == 0, lax._const(d, 0),
+               lax.div(lax.abs(multiply(x1, x2)), d))
+
+
 ### track unimplemented functions
 
 def _not_implemented(fun):
@@ -2094,8 +2341,8 @@ _nondiff_methods = ["all", "any", "argmax", "argmin", "argpartition", "argsort",
                     "nonzero", "searchsorted", "round"]
 _diff_methods = ["clip", "compress", "conj", "conjugate", "cumprod", "cumsum",
                  "diagonal", "dot", "max", "mean", "min", "prod", "ptp",
-                 "ravel", "repeat", "reshape", "sort", "squeeze", "std", "sum",
-                 "swapaxes", "take", "trace", "transpose", "var"]
+                 "ravel", "repeat", "sort", "squeeze", "std", "sum",
+                 "swapaxes", "take", "tile", "trace", "transpose", "var"]
 
 
 # Set up operator, method, and property forwarding on Tracer instances containing
@@ -2106,20 +2353,25 @@ for operator_name, function in _operators.items():
 # Forward methods and properties using core.aval_method and core.aval_property:
 for method_name in _nondiff_methods + _diff_methods:
   setattr(ShapedArray, method_name, core.aval_method(globals()[method_name]))
+setattr(ShapedArray, "reshape", core.aval_method(_reshape))
 setattr(ShapedArray, "flatten", core.aval_method(ravel))
 setattr(ShapedArray, "T", core.aval_property(transpose))
+setattr(ShapedArray, "real", core.aval_property(real))
+setattr(ShapedArray, "imag", core.aval_property(imag))
 setattr(ShapedArray, "astype", core.aval_method(lax.convert_element_type))
-setattr(ShapedArray, "psum", core.aval_method(lax.psum))
 
 
-# Forward operators, methods, and properies on DeviceArray to lax_numpy
+# Forward operators, methods, and properties on DeviceArray to lax_numpy
 # functions (with no Tracers involved; this forwarding is direct)
 for operator_name, function in _operators.items():
   setattr(DeviceArray, "__{}__".format(operator_name), function)
 for method_name in _nondiff_methods + _diff_methods:
   setattr(DeviceArray, method_name, globals()[method_name])
+setattr(DeviceArray, "reshape", _reshape)
 setattr(DeviceArray, "flatten", ravel)
 setattr(DeviceArray, "T", property(transpose))
+setattr(DeviceArray, "real", property(real))
+setattr(DeviceArray, "imag", property(imag))
 setattr(DeviceArray, "astype", lax.convert_element_type)
 
 
