@@ -63,7 +63,7 @@ def svd(a, full_matrices=True, compute_uv=True):
 @_wraps(onp.linalg.slogdet)
 def slogdet(a):
   a = _promote_arg_dtypes(np.asarray(a))
-  dtype = lax._dtype(a)
+  dtype = lax.dtype(a)
   a_shape = np.shape(a)
   if len(a_shape) < 2 or a_shape[-1] != a_shape[-2]:
     msg = "Argument to slogdet() must have shape [..., n, n], got {}"
@@ -92,8 +92,15 @@ def det(a):
   return sign * np.exp(logdet)
 
 
+@_wraps(onp.linalg.eig)
+def eig(a):
+  a = _promote_arg_dtypes(np.asarray(a))
+  w, vl, vr = lax_linalg.eig(a)
+  return w, vr
+
+
 @_wraps(onp.linalg.eigh)
-def eigh(a, UPLO=None):
+def eigh(a, UPLO=None, symmetrize_input=True):
   if UPLO is None or UPLO == "L":
     lower = True
   elif UPLO == "U":
@@ -103,7 +110,7 @@ def eigh(a, UPLO=None):
     raise ValueError(msg)
 
   a = _promote_arg_dtypes(np.asarray(a))
-  v, w = lax_linalg.eigh(a, lower=lower)
+  v, w = lax_linalg.eigh(a, lower=lower, symmetrize_input=symmetrize_input)
   return w, v
 
 
@@ -140,7 +147,7 @@ def norm(x, ord=None, axis=None, keepdims=False):
     elif ord == -np.inf:
       return np.amin(np.abs(x), axis=axis, keepdims=keepdims)
     elif ord == 0:
-      return np.sum(x != 0, dtype=np.finfo(lax._dtype(x)).dtype,
+      return np.sum(x != 0, dtype=np.finfo(lax.dtype(x)).dtype,
                     axis=axis, keepdims=keepdims)
     elif ord == 1:
       # Numpy has a special case for ord == 1 as an optimization. We don't
@@ -228,7 +235,7 @@ def solve(a, b):
            "b=[..., m, k] or b=[..., m]; got a={} and b={}")
     raise ValueError(msg.format(a_shape, b_shape))
   lu, pivots = lax_linalg.lu(a)
-  dtype = lax._dtype(a)
+  dtype = lax.dtype(a)
 
   # TODO(phawkins): add unit_diagonal support to solve_triangular, use it here
   # instead of explicit masking of l.
