@@ -16,7 +16,6 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-import numpy as onp
 import scipy.stats as osp_stats
 
 from ... import lax
@@ -26,10 +25,9 @@ from ...numpy.lax_numpy import _promote_args_like, _wraps, where, inf, logical_o
 @_wraps(osp_stats.uniform.logpdf)
 def logpdf(x, loc=0, scale=1):
   x, loc, scale = _promote_args_like(osp_stats.uniform.logpdf, x, loc, scale)
-  fill_value = lax.neg(lax.log(scale))
-  log_probs = lax.broadcast(fill_value, onp.shape(x))
-  return where(logical_or(lax.ge(x, lax.add(loc, scale)),
-                          lax.le(x, loc)), -inf, log_probs)
+  log_probs = lax.neg(lax.log(scale))
+  return where(logical_or(lax.gt(x, lax.add(loc, scale)),
+                          lax.lt(x, loc)), -inf, log_probs)
 
 @_wraps(osp_stats.uniform.pdf)
 def pdf(x, loc=0, scale=1):
