@@ -20,7 +20,7 @@ import functools
 import re
 import itertools as it
 import os
-import random
+from unittest import SkipTest
 
 from absl.testing import absltest
 from absl.testing import parameterized
@@ -34,22 +34,22 @@ from . import api
 from .config import flags
 from .util import partial
 from .tree_util import tree_multimap, tree_all, tree_map, tree_reduce
+from .lib import xla_bridge
 
 # lbr tests placeholder
 
 
 FLAGS = flags.FLAGS
 flags.DEFINE_enum(
-    'jax_test_dut',
-    'cpu',
-    enum_values=['cpu', 'gpu', 'tpu'],
+    'jax_test_dut', '',
+    enum_values=['', 'cpu', 'gpu', 'tpu'],
     help=
     'Describes the device under test in case special consideration is required.'
 )
 
 flags.DEFINE_integer(
   'num_generated_cases',
-  os.getenv('JAX_NUM_GENERATED_CASES', 10),
+  int(os.getenv('JAX_NUM_GENERATED_CASES', 10)),
   help='Number of generated cases to test')
 
 EPS = 1e-4
@@ -59,6 +59,14 @@ RTOL = 1e-4
 _dtype = lambda x: getattr(x, 'dtype', None) or onp.asarray(x).dtype
 
 
+def is_sequence(x):
+  try:
+    iter(x)
+  except TypeError:
+    return False
+  else:
+    return True
+
 def numpy_eq(x, y):
   testing_tpu = FLAGS.jax_test_dut and FLAGS.jax_test_dut.startswith("tpu")
   testing_x32 = not FLAGS.jax_enable_x64
@@ -114,10 +122,10 @@ def rand_like(rng, x):
 
 
 def numerical_jvp(f, primals, tangents, eps=EPS):
-  delta = scalar_mul(tangents, EPS)
+  delta = scalar_mul(tangents, eps)
   f_pos = f(*add(primals, delta))
   f_neg = f(*sub(primals, delta))
-  return scalar_mul(sub(f_pos, f_neg), 0.5 / EPS)
+  return scalar_mul(sub(f_pos, f_neg), 0.5 / eps)
 
 
 def check_jvp(f, f_jvp, args, atol=ATOL, rtol=RTOL, eps=EPS):
@@ -136,7 +144,7 @@ def check_vjp(f, f_vjp, args, atol=ATOL, rtol=RTOL, eps=EPS):
   v_out_expected = f(*args)
   check_eq(v_out, v_out_expected)
   tangent = tree_map(_rand_like, args)
-  tangent_out = numerical_jvp(f, args, tangent, eps=EPS)
+  tangent_out = numerical_jvp(f, args, tangent, eps=eps)
   cotangent = tree_map(_rand_like, v_out)
   cotangent_out = conj(vjpfun(conj(cotangent)))
   ip = inner_prod(tangent, cotangent_out)
@@ -144,20 +152,32 @@ def check_vjp(f, f_vjp, args, atol=ATOL, rtol=RTOL, eps=EPS):
   check_close(ip, ip_expected, atol=atol, rtol=rtol)
 
 
-def check_grads(f, args, order, atol=None, rtol=None, eps=None):
-  if order > 1:
-    def f_vjp(*args):
-      out_primal_py, vjp_py = api.vjp(f, *args)
-      return vjp_py(out_primal_py)
+def check_grads(f, args, order,
+                modes=["fwd", "rev"], atol=None, rtol=None, eps=None):
+  args = tuple(args)
+  default_tol = 1e-6 if FLAGS.jax_enable_x64 else 1e-2
+  atol = atol or default_tol
+  rtol = rtol or default_tol
+  eps = eps or EPS
 
-    check_grads(f_vjp, args, order - 1, atol=atol, rtol=rtol, eps=eps)
-  else:
-    default_tol = 1e-6 if FLAGS.jax_enable_x64 else 1e-2
-    atol = atol or default_tol
-    rtol = rtol or default_tol
-    eps = eps or default_tol
-    check_jvp(f, partial(api.jvp, f), args, atol, rtol, eps)
-    check_vjp(f, partial(api.vjp, f), args, atol, rtol, eps)
+  _check_jvp = partial(check_jvp, atol=atol, rtol=rtol, eps=eps)
+  _check_vjp = partial(check_vjp, atol=atol, rtol=rtol, eps=eps)
+
+  def _check_grads(f, args, order):
+    if "fwd" in modes:
+      _check_jvp(f, partial(api.jvp, f), args)
+      if order > 1:
+        _check_grads(partial(api.jvp, f), (args, args), order - 1)
+
+    if "rev" in modes:
+      _check_vjp(f, partial(api.vjp, f), args)
+      if order > 1:
+        def f_vjp(*args):
+          out_primal_py, vjp_py = api.vjp(f, *args)
+          return vjp_py(out_primal_py)
+        _check_grads(f_vjp, args, order - 1)
+
+  _check_grads(f, args, order)
 
 
 def skip_on_devices(*disabled_devices):
@@ -165,11 +185,11 @@ def skip_on_devices(*disabled_devices):
   def skip(test_method):
     @functools.wraps(test_method)
     def test_method_wrapper(self, *args, **kwargs):
-      device = FLAGS.jax_test_dut
+      device = FLAGS.jax_test_dut or xla_bridge.get_backend().platform
       if device in disabled_devices:
         test_name = getattr(test_method, '__name__', '[unknown test]')
-        return absltest.unittest.skip(
-            '{} not supported on {}.'.format(test_name, device.upper()))
+        raise SkipTest('{} not supported on {}.'
+                       .format(test_name, device.upper()))
       return test_method(self, *args, **kwargs)
     return test_method_wrapper
   return skip
@@ -183,9 +203,8 @@ def skip_on_flag(flag_name, skip_value):
       flag_value = getattr(FLAGS, flag_name)
       if flag_value == skip_value:
         test_name = getattr(test_method, '__name__', '[unknown test]')
-        return absltest.unittest.skip(
-            '{} not supported when FLAGS.{} is {}'.format(
-                test_name, flag_name, flag_value))
+        raise SkipTest('{} not supported when FLAGS.{} is {}'
+                       .format(test_name, flag_name, flag_value))
       return test_method(self, *args, **kwargs)
     return test_method_wrapper
   return skip
@@ -197,41 +216,57 @@ def format_test_name_suffix(opname, shapes, dtypes):
   return '{}_{}'.format(opname.capitalize(), '_'.join(arg_descriptions))
 
 
-class _NumpyScalar(object):
-
-  def __len__(self):
-    return 0
-
-# A special singleton "shape" that denotes numpy scalars. Numpy scalars are not
-# identical to 0-D arrays, and we want to write tests that exercise both paths.
+# We use special symbols, represented as singleton objects, to distinguish
+# between NumPy scalars, Python scalars, and 0-D arrays.
+class ScalarShape(object):
+  def __len__(self): return 0
+class _NumpyScalar(ScalarShape): pass
+class _PythonScalar(ScalarShape): pass
 NUMPY_SCALAR_SHAPE = _NumpyScalar()
+PYTHON_SCALAR_SHAPE = _PythonScalar()
 
 
 def _dims_of_shape(shape):
   """Converts `shape` to a tuple of dimensions."""
-  return shape if shape != NUMPY_SCALAR_SHAPE else ()
+  if type(shape) in (list, tuple):
+    return shape
+  elif isinstance(shape, ScalarShape):
+    return ()
+  else:
+    raise TypeError(type(shape))
 
 
 def _cast_to_shape(value, shape, dtype):
   """Casts `value` to the correct Python type for `shape` and `dtype`."""
-  if shape != NUMPY_SCALAR_SHAPE:
+  if shape is NUMPY_SCALAR_SHAPE:
+    # explicitly cast to NumPy scalar in case `value` is a Python scalar.
+    return dtype(value)
+  elif shape is PYTHON_SCALAR_SHAPE:
+    # explicitly cast to Python scalar via https://stackoverflow.com/a/11389998
+    return onp.asarray(value).item()
+  elif type(shape) in (list, tuple):
+    assert onp.shape(value) == tuple(shape)
     return value
   else:
-    # A numpy scalar was requested. Explicitly cast in case `value` is a Python
-    # scalar.
-    return dtype(value)
+    raise TypeError(type(shape))
+
+
+def dtype_str(dtype):
+  return onp.dtype(dtype).name
 
 
 def format_shape_dtype_string(shape, dtype):
-  typestr = onp.dtype(dtype).name
-  if shape == NUMPY_SCALAR_SHAPE:
-    return typestr
-
-  if onp.isscalar(shape):
-    shapestr = str(shape) + ','
-  else:
+  if shape is NUMPY_SCALAR_SHAPE:
+    return dtype_str(dtype)
+  elif shape is PYTHON_SCALAR_SHAPE:
+    return 'py' + dtype_str(dtype)
+  elif type(shape) in (list, tuple):
     shapestr = ','.join(str(dim) for dim in shape)
-  return '{}[{}]'.format(typestr, shapestr)
+    return '{}[{}]'.format(dtype_str(dtype), shapestr)
+  elif type(shape) is int:
+    return '{}[{},]'.format(dtype_str(dtype), shape)
+  else:
+    raise TypeError(type(shape))
 
 
 def _rand_dtype(rand, shape, dtype, scale=1., post=lambda x: x):
@@ -311,7 +346,6 @@ def rand_some_equal():
   return partial(_rand_dtype, randn, scale=100., post=post)
 
 
-# TODO(mattjj): doesn't handle complex types
 def rand_some_inf():
   """Return a random sampler that produces infinities in floating types."""
   rng = npr.RandomState(1)
@@ -323,6 +357,10 @@ def rand_some_inf():
       # only float types have inf
       return base_rand(shape, dtype)
 
+    if onp.issubdtype(dtype, onp.complexfloating):
+      base_dtype = onp.real(onp.array(0, dtype=dtype)).dtype
+      return rand(shape, base_dtype) + 1j * rand(shape, base_dtype)
+
     dims = _dims_of_shape(shape)
     posinf_flips = rng.rand(*dims) < 0.1
     neginf_flips = rng.rand(*dims) < 0.1
@@ -335,6 +373,34 @@ def rand_some_inf():
 
   return rand
 
+def rand_some_inf_and_nan():
+  """Return a random sampler that produces infinities in floating types."""
+  rng = npr.RandomState(1)
+  base_rand = rand_default()
+
+  def rand(shape, dtype):
+    """The random sampler function."""
+    if not onp.issubdtype(dtype, onp.floating):
+      # only float types have inf
+      return base_rand(shape, dtype)
+
+    if onp.issubdtype(dtype, onp.complexfloating):
+      base_dtype = onp.real(onp.array(0, dtype=dtype)).dtype
+      return rand(shape, base_dtype) + 1j * rand(shape, base_dtype)
+
+    dims = _dims_of_shape(shape)
+    posinf_flips = rng.rand(*dims) < 0.1
+    neginf_flips = rng.rand(*dims) < 0.1
+    nan_flips = rng.rand(*dims) < 0.1
+
+    vals = base_rand(shape, dtype)
+    vals = onp.where(posinf_flips, onp.inf, vals)
+    vals = onp.where(neginf_flips, -onp.inf, vals)
+    vals = onp.where(nan_flips, onp.nan, vals)
+
+    return _cast_to_shape(onp.asarray(vals, dtype=dtype), shape, dtype)
+
+  return rand
 
 # TODO(mattjj): doesn't handle complex types
 def rand_some_zero():
@@ -381,12 +447,12 @@ def check_raises_regexp(thunk, err_type, pattern):
   except err_type as e:
     assert re.match(pattern, str(e)), "{}\n\n{}\n".format(e, pattern)
 
-random.seed(0) # TODO: consider managing prng state more carefully
-
 def cases_from_list(xs):
+  rng = npr.RandomState(42)
   xs = list(xs)
   k = min(len(xs), FLAGS.num_generated_cases)
-  return random.sample(xs, k)
+  indices = rng.choice(onp.arange(len(xs)), k, replace=False)
+  return [xs[i] for i in indices]
 
 def cases_from_gens(*gens):
   sizes = [1, 3, 10]
@@ -401,6 +467,7 @@ class JaxTestCase(parameterized.TestCase):
 
   def assertArraysAllClose(self, x, y, check_dtypes, atol=None, rtol=None):
     """Assert that x and y are close (up to numerical tolerances)."""
+    self.assertEqual(x.shape, y.shape)
     dtype = lambda x: str(onp.asarray(x).dtype)
     tol = 1e-2 if str(onp.dtype(onp.float32)) in {dtype(x), dtype(y)} else 1e-5
     atol = atol or tol
@@ -434,23 +501,23 @@ class JaxTestCase(parameterized.TestCase):
 
   def assertAllClose(self, x, y, check_dtypes, atol=None, rtol=None):
     """Assert that x and y, either arrays or nested tuples/lists, are close."""
-    if isinstance(x, (tuple, list)):
-      self.assertIsInstance(y, (tuple, list))
-      self.assertEqual(len(x), len(y))
-      for x_elt, y_elt in zip(x, y):
-        self.assertAllClose(x_elt, y_elt, check_dtypes, atol=atol, rtol=rtol)
-    elif isinstance(x, dict):
+    if isinstance(x, dict):
       self.assertIsInstance(y, dict)
       self.assertEqual(set(x.keys()), set(y.keys()))
       for k in x.keys():
         self.assertAllClose(x[k], y[k], check_dtypes, atol=atol, rtol=rtol)
-    else:
-      is_array = lambda x: hasattr(x, '__array__') or onp.isscalar(x)
-      self.assertTrue(is_array(x))
-      self.assertTrue(is_array(y))
+    elif is_sequence(x) and not hasattr(x, '__array__'):
+      self.assertTrue(is_sequence(y) and not hasattr(y, '__array__'))
+      self.assertEqual(len(x), len(y))
+      for x_elt, y_elt in zip(x, y):
+        self.assertAllClose(x_elt, y_elt, check_dtypes, atol=atol, rtol=rtol)
+    elif hasattr(x, '__array__') or onp.isscalar(x):
+      self.assertTrue(hasattr(y, '__array__') or onp.isscalar(y))
       x = onp.asarray(x)
       y = onp.asarray(y)
       self.assertArraysAllClose(x, y, check_dtypes, atol=atol, rtol=rtol)
+    else:
+      raise TypeError((type(x), type(y)))
 
   def _CompileAndCheck(self, fun, args_maker, check_dtypes,
                        rtol=None, atol=None):
@@ -484,7 +551,7 @@ class JaxTestCase(parameterized.TestCase):
 
     self.assertAllClose(python_ans, compiled_ans, check_dtypes, rtol, atol)
 
-  def _CheckAgainstNumpy(self, lax_op, numpy_reference_op, args_maker,
+  def _CheckAgainstNumpy(self, numpy_reference_op, lax_op, args_maker,
                          check_dtypes=False, tol=1e-5):
     args = args_maker()
     numpy_ans = numpy_reference_op(*args)
