@@ -12,6 +12,25 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+"""Utilities for working with tree-like container data structures.
+
+The code here is independent of JAX. The only dependence is on jax.util, which
+itself has no JAX-specific code.
+
+This module provides a small set of utility functions for working with tree-like
+data structures, such as nested tuples, lists, and dicts. We call these
+structures pytrees. They are trees in that they are defined recursively (any
+non-pytree is a pytree, i.e. a leaf, and any pytree of pytrees is a pytree) and
+can be operated on recursively (object identity equivalence is not preserved by
+mapping operations, and the structures cannot contain reference cycles).
+
+The set of Python types that are considered pytree nodes (e.g. that can be
+mapped over, rather than treated as leaves) is extensible. There is a single
+module-level registry of types, and class hierarchy is ignored. By registering a
+new pytree node type, that type in effect becomes transparent to the utility
+functions in this file.
+"""
+
 from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
@@ -26,7 +45,18 @@ map = safe_map
 
 
 def tree_map(f, tree):
-  node_type = node_types.get(type(tree))
+  """Map a function over a pytree to produce a new pytree.
+
+  Args:
+    f: function to be applied at each leaf.
+    tree: a pytree to be mapped over.
+
+  Returns:
+    A new pytree with the same structure as `tree` but with the value at each
+    leaf given by `f(x)` where `x` is the value at the corresponding leaf in
+    `tree`.
+  """
+  node_type = _get_node_type(tree)
   if node_type:
     children, node_spec = node_type.to_iterable(tree)
     new_children = [tree_map(f, child) for child in children]
@@ -34,21 +64,36 @@ def tree_map(f, tree):
   else:
     return f(tree)
 
-
 def tree_multimap(f, tree, *rest):
-  tree_type = type(tree)
-  node_type = node_types.get(tree_type)
+  """Map a multi-input function over pytree args to produce a new pytree.
+
+  Args:
+    f: function that takes `1 + len(rest)` arguments, to be applied at the
+      corresponding leaves of the pytrees.
+    tree: a pytree to be mapped over, with each leaf providing the first
+      positional argument to `f`.
+    *rest: a tuple of pytrees, each with the same structure as `tree`.
+
+  Returns:
+    A new pytree with the same structure as `tree` but with the value at each
+    leaf given by `f(x, *xs)` where `x` is the value at the corresponding leaf
+    in `tree` and `xs` is the tuple of values at corresponding leaves in `rest`.
+  """
+  node_type = _get_node_type(tree)
   if node_type:
-    children, node_spec = node_type.to_iterable(tree)
+    children, aux_data = node_type.to_iterable(tree)
     all_children = [children]
     for other_tree in rest:
-      other_children, other_node_spec = node_type.to_iterable(other_tree)
-      if other_node_spec != node_spec:
-        raise TypeError('Mismatch: {} != {}'.format(other_node_spec, node_spec))
+      other_node_type = _get_node_type(other_tree)
+      if node_type != other_node_type:
+        raise TypeError('Mismatch: {} != {}'.format(other_node_type, node_type))
+      other_children, other_aux_data = node_type.to_iterable(other_tree)
+      if other_aux_data != aux_data:
+        raise TypeError('Mismatch: {} != {}'.format(other_aux_data, aux_data))
       all_children.append(other_children)
 
     new_children = [tree_multimap(f, *xs) for xs in zip(*all_children)]
-    return node_type.from_iterable(node_spec, new_children)
+    return node_type.from_iterable(aux_data, new_children)
   else:
     return f(tree, *rest)
 
@@ -68,7 +113,7 @@ def process_pytree(process_node, tree):
 
 
 def walk_pytree(f_node, f_leaf, tree):
-  node_type = node_types.get(type(tree))
+  node_type = _get_node_type(tree)
   if node_type:
     children, node_spec = node_type.to_iterable(tree)
     proc_children, child_specs = unzip2([walk_pytree(f_node, f_leaf, child)
@@ -91,7 +136,6 @@ def build_tree(treedef, xs):
 tree_flatten = partial(walk_pytree, concatenate, lambda x: [x])
 
 def tree_unflatten(treedef, xs):
-  # like build_tree, but handles empty containers in the tree
   return _tree_unflatten(iter(xs), treedef)
 
 def _tree_unflatten(xs, treedef):
@@ -133,15 +177,6 @@ def tree_structure(tree):
   return spec
 
 
-def prune(treedef, tuple_tree):
-  if treedef is leaf:
-    return tuple_tree
-  elif treedef.children:
-    return tuple(map(prune, treedef.children, tuple_tree))
-  else:
-    return ()
-
-
 class PyTreeDef(object):
   def __init__(self, node_type, node_data, children):
     self.node_type = node_type
@@ -188,6 +223,9 @@ class NodeType(object):
     self.to_iterable = to_iterable
     self.from_iterable = from_iterable
 
+  def __repr__(self):
+    return self.name
+
 node_types = {}
 
 def register_pytree_node(py_type, to_iterable, from_iterable):
@@ -198,3 +236,20 @@ register_pytree_node(tuple, lambda xs: (xs, None), lambda _, xs: tuple(xs))
 register_pytree_node(list, lambda xs: (tuple(xs), None), lambda _, xs: list(xs))
 register_pytree_node(dict, dict_to_iterable, lambda keys, xs: dict(zip(keys, xs)))
 register_pytree_node(type(None), lambda z: ((), None), lambda _, xs: None)
+
+
+# To handle namedtuples, we can't just use the standard table of node_types
+# because every namedtuple creates its own type and thus would require its own
+# entry in the table. Instead we use a heuristic check on the type itself to
+# decide whether it's a namedtuple type, and if so treat it as a pytree node.
+def _get_node_type(maybe_tree):
+  t = type(maybe_tree)
+  return node_types.get(t) or _namedtuple_node(t)
+
+def _namedtuple_node(t):
+  if t.__bases__ == (tuple,) and hasattr(t, '_fields'):
+    return NamedtupleNode
+
+NamedtupleNode = NodeType('namedtuple',
+                          lambda xs: (tuple(xs), type(xs)),
+                          lambda t, xs: t(*xs))
